from .. import Provider as DateTimeProvider


class Provider(DateTimeProvider):

    DAY_NAMES = {
        "0": "Воскресенье",
        "1": "Понедельник",
        "2": "Вторник",
        "3": "Среда",
        "4": "Четверг",
        "5": "Пятница",
        "6": "Суббота",
    }

    MONTH_NAMES = {
        "01": "Январь",
        "02": "Февраль",
        "03": "Март",
        "04": "Апрель",
        "05": "Май",
        "06": "Июнь",
        "07": "Июль",
        "08": "Август",
        "09": "Сентябрь",
        "10": "Октябрь",
        "11": "Ноябрь",
        "12": "Декабрь",
    }

    # Timezone names are based on Wiki list, source: https://ru.wikipedia.org/wiki/Список_часовых_поясов_по_странам
    countries = [{'timezones': ['Андорра (UTC+01)'],
                  'alpha-2-code': 'AD',
                  'alpha-3-code': 'AND',
                  'continent': 'Европа',
                  'name': 'Андорра',
                  'capital': 'Андорра-ла-Велья'},
                 {'timezones': ['Афганистан (UTC+04:30)'],
                  'alpha-2-code': 'AF',
                  'alpha-3-code': 'AFG',
                  'continent': 'Азия',
                  'name': 'Афганистан',
                  'capital': 'Кабул'},
                 {'timezones': ['Антигуа и Барбуда (UTC-04)'],
                  'alpha-2-code': 'AG',
                  'alpha-3-code': 'ATG',
                  'continent': 'Северная Америка',
                  'name': 'Антигуа и Барбуда',
                  'capital': "Сент-Джонс"},
                 {'timezones': ['Албания (UTC+01)'],
                  'alpha-2-code': 'AL',
                  'alpha-3-code': 'ALB',
                  'continent': 'Европа',
                  'name': 'Албания',
                  'capital': 'Тирана'},
                 {'timezones': ['Армения (UTC+04)'],
                  'alpha-2-code': 'AM',
                  'alpha-3-code': 'ARM',
                  'continent': 'Азия',
                  'name': 'Армения',
                  'capital': 'Ереван'},
                 {'timezones': ['Ангола (UTC+01)'],
                  'alpha-2-code': 'AO',
                  'alpha-3-code': 'AGO',
                  'continent': 'Африка',
                  'name': 'Ангола',
                  'capital': 'Луанда'},
                 {'timezones': ['Аргентина (UTC-03)'],
                  'alpha-2-code': 'AR',
                  'alpha-3-code': 'ARG',
                  'continent': 'Южная Америка',
                  'name': 'Аргентина',
                  'capital': 'Буэнос Айрес'},
                 {'timezones': ['Австрия (UTC+01)'],
                  'alpha-2-code': 'AT',
                  'alpha-3-code': 'AUT',
                  'continent': 'Европа',
                  'name': 'Австрия',
                  'capital': 'Вена'},
                 {'timezones': ['Австралия (UTC+05)',
                                'Австралия (UTC+06:30)',
                                'Австралия (UTC+07)',
                                'Австралия (UTC+08)',
                                'Австралия (UTC+9:30)',
                                'Австралия (UTC+10)',
                                'Австралия (UTC+10:30)',
                                'Австралия (UTC+11:30)'],
                  'alpha-2-code': 'AU',
                  'alpha-3-code': 'AUS',
                  'continent': 'Океания',
                  'name': 'Австралия',
                  'capital': 'Канберра'},
                 {'timezones': ['Азербайджан (UTC+04)'],
                  'alpha-2-code': 'AZ',
                  'alpha-3-code': 'AZE',
                  'continent': 'Азия',
                  'name': 'Азербайджан',
                  'capital': 'Баку'},
                 {'timezones': ['Барбадос (UTC-04)'],
                  'alpha-2-code': 'BB',
                  'alpha-3-code': 'BRB',
                  'continent': 'Северная Америка',
                  'name': 'Барбадос',
                  'capital': 'Бриджтаун'},
                 {'timezones': ['Бангладеш (UTC+06)'],
                  'alpha-2-code': 'BD',
                  'alpha-3-code': 'BGD',
                  'continent': 'Азия',
                  'name': 'Бангладеш',
                  'capital': 'Дака'},
                 {'timezones': ['Бельгия (UTC+01)'],
                  'alpha-2-code': 'BE',
                  'alpha-3-code': 'BEL',
                  'continent': 'Европа',
                  'name': 'Бельгия',
                  'capital': 'Брюссель'},
                 {'timezones': ['Буркина-Фасо (UTC)'],
                  'alpha-2-code': 'BF',
                  'alpha-3-code': 'BFA',
                  'continent': 'Африка',
                  'name': 'Буркина-Фасо',
                  'capital': 'Уагадугу'},
                 {'timezones': ['Болгария (UTC+02)'],
                  'alpha-2-code': 'BG',
                  'alpha-3-code': 'BGR',
                  'continent': 'Европа',
                  'name': 'Болгария',
                  'capital': 'София'},
                 {'timezones': ['Бахрейн (UTC+03)'],
                  'alpha-2-code': 'BH',
                  'alpha-3-code': 'BHR',
                  'continent': 'Азия',
                  'name': 'Бахрейн',
                  'capital': 'Манама'},
                 {'timezones': ['Бурунди (UTC+02)'],
                  'alpha-2-code': 'BI',
                  'alpha-3-code': 'BDI',
                  'continent': 'Африка',
                  'name': 'Бурунди',
                  'capital': 'Гитега'},
                 {'timezones': ['Бенин (UTC+01)'],
                  'alpha-2-code': 'BJ',
                  'alpha-3-code': 'BEN',
                  'continent': 'Африка',
                  'name': 'Бенин',
                  'capital': 'Порто-Ново'},
                 {'timezones': ['Бруней (UTC+08)'],
                  'alpha-2-code': 'BN',
                  'alpha-3-code': 'BRN',
                  'continent': 'Азия',
                  'name': 'Бруней',
                  'capital': 'Бандар-Сери-Бегаван'},
                 {'timezones': ['Боливия (UTC-04)'],
                  'alpha-2-code': 'BO',
                  'alpha-3-code': 'BOL',
                  'continent': 'Южная Америка',
                  'name': 'Боливия',
                  'capital': 'Сукре'},
                 {'timezones': ['Бразилия (UTC-05)',
                                'Бразилия (UTC-04)',
                                'Бразилия (UTC-03)',
                                'Бразилия (UTC-02)'],
                  'alpha-2-code': 'BR',
                  'alpha-3-code': 'BRA',
                  'continent': 'Южная Америка',
                  'name': 'Бразилия',
                  'capital': 'Бразилиа'},
                 {'timezones': ['Багамские Острова (UTC-05)'],
                  'alpha-2-code': 'BS',
                  'alpha-3-code': 'BHS',
                  'continent': 'Северная Америка',
                  'name': 'Багамские Острова',
                  'capital': 'Нассау'},
                 {'timezones': ['Бутан (UTC+06)'],
                  'alpha-2-code': 'BT',
                  'alpha-3-code': 'BTN',
                  'continent': 'Азия',
                  'name': 'Бутан',
                  'capital': 'Тхимпху'},
                 {'timezones': ['Ботсвана (UTC+02)'],
                  'alpha-2-code': 'BW',
                  'alpha-3-code': 'BWA',
                  'continent': 'Африка',
                  'name': 'Ботсвана',
                  'capital': 'Габороне'},
                 {'timezones': ['Белоруссия (UTC+03)'],
                  'alpha-2-code': 'BY',
                  'alpha-3-code': 'BLR',
                  'continent': 'Европа',
                  'name': 'Белоруссия',
                  'capital': 'Минск'},
                 {'timezones': ['Белиз (UTC-06)'],
                  'alpha-2-code': 'BZ',
                  'alpha-3-code': 'BLZ',
                  'continent': 'Северная Америка',
                  'name': 'Белиз',
                  'capital': 'Бельмопан'},
                 {'timezones': ['Канада (UTC-08)',
                                'Канада (UTC-07)',
                                'Канада (UTC-06)',
                                'Канада (UTC-05)',
                                'Канада (UTC-04)',
                                'Канада (UTC-03:30)'],
                  'alpha-2-code': 'CA',
                  'alpha-3-code': 'CAN',
                  'continent': 'Северная Америка',
                  'name': 'Канада',
                  'capital': 'Оттава'},
                 {'timezones': ['Демократическая Республика Конго (UTC+01)',
                                'Демократическая Республика Конго (UTC+02)'],
                  'alpha-2-code': 'CD',
                  'alpha-3-code': 'COD',
                  'continent': 'Африка',
                  'name': 'Демократическая Республика Конго',
                  'capital': 'Киншаса'},
                 {'timezones': ['Республика Конго (UTC+01)'],
                  'alpha-2-code': 'CG',
                  'alpha-3-code': 'COG',
                  'continent': 'Африка',
                  'name': 'Руспублика Конго',
                  'capital': 'Браззавиль'},
                 {'timezones': ["Кот-д'Ивуар (UTC)"],
                  'alpha-2-code': 'CI',
                  'alpha-3-code': 'CIV',
                  'continent': 'Африка',
                  'name': "Кот-д'Ивуар",
                  'capital': 'Ямусукро'},
                 {'timezones': ['Чили (UTC-06)',
                                'Чили (UTC-04)'],
                  'alpha-2-code': 'CL',
                  'alpha-3-code': 'CHL',
                  'continent': 'Южная Америка',
                  'name': 'Чили',
                  'capital': 'Сантьяго'},
                 {'timezones': ['Камерун (UTC+01)'],
                  'alpha-2-code': 'CM',
                  'alpha-3-code': 'CMR',
                  'continent': 'Африка',
                  'name': 'Камерун',
                  'capital': 'Яунде'},
                 {'timezones': ['Китай (UTC+08)'],
                  'alpha-2-code': 'CN',
                  'alpha-3-code': 'CHN',
                  'continent': 'Азия',
                  'name': "Китайская Народная Республика",
                  'capital': 'Пекин'},
                 {'timezones': ['Колумбия (UTC-05)'],
                  'alpha-2-code': 'CO',
                  'alpha-3-code': 'COL',
                  'continent': 'Южная Америка',
                  'name': 'Колумбия',
                  'capital': 'Богота'},
                 {'timezones': ['Коста-Рика (UTC-06)'],
                  'alpha-2-code': 'CR',
                  'alpha-3-code': 'CRI',
                  'continent': 'Северная Америка',
                  'name': 'Коста-Рика',
                  'capital': 'Сан-Хосе'},
                 {'timezones': ['Куба (UTC-05)'],
                  'alpha-2-code': 'CU',
                  'alpha-3-code': 'CUB',
                  'continent': 'Северная Америка',
                  'name': 'Куба',
                  'capital': 'Гавана'},
                 {'timezones': ['Кабо-Верде (UTC-01)'],
                  'alpha-2-code': 'CV',
                  'alpha-3-code': 'CPV',
                  'continent': 'Африка',
                  'name': 'Кабо-Верде',
                  'capital': 'Прая'},
                 {'timezones': ['Кипр (UTC+02)'],
                  'alpha-2-code': 'CY',
                  'alpha-3-code': 'CYP',
                  'continent': 'Азия',
                  'name': 'Кипр',
                  'capital': 'Никосия'},
                 {'timezones': ['Чехия (UTC+01)'],
                  'alpha-2-code': 'CZ',
                  'alpha-3-code': 'CZE',
                  'continent': 'Европа',
                  'name': 'Чехия',
                  'capital': 'Прага'},
                 {'timezones': ['Германия (UTC+01)'],
                  'alpha-2-code': 'DE',
                  'alpha-3-code': 'DEU',
                  'continent': 'Европа',
                  'name': 'Германия',
                  'capital': 'Берлин'},
                 {'timezones': ['Джибути (UTC+03)'],
                  'alpha-2-code': 'DJ',
                  'alpha-3-code': 'DJI',
                  'continent': 'Африка',
                  'name': 'Джибути',
                  'capital': 'Джибути'},
                 {'timezones': ['Дания (UTC+01)'],
                  'alpha-2-code': 'DK',
                  'alpha-3-code': 'DNK',
                  'continent': 'Европа',
                  'name': 'Дания',
                  'capital': 'Копенгаген'},
                 {'timezones': ['Доминика (UTC-04)'],
                  'alpha-2-code': 'DM',
                  'alpha-3-code': 'DMA',
                  'continent': 'Северная Америка',
                  'name': 'Доминика',
                  'capital': 'Розо'},
                 {'timezones': ['Доминиканская Республика (UTC-04)'],
                  'alpha-2-code': 'DO',
                  'alpha-3-code': 'DOM',
                  'continent': 'Северная Америка',
                  'name': 'Доминиканская Республика',
                  'capital': 'Санто-Доминго'},
                 {'timezones': ['Эквадор (UTC-06)',
                                'Эквадор (UTC-05)'],
                  'alpha-2-code': 'EC',
                  'alpha-3-code': 'ECU',
                  'continent': 'Южная Америка',
                  'name': 'Эквадор',
                  'capital': 'Кито'},
                 {'timezones': ['Эстония (UTC+02)'],
                  'alpha-2-code': 'EE',
                  'alpha-3-code': 'EST',
                  'continent': 'Европа',
                  'name': 'Эстония',
                  'capital': 'Таллинн'},
                 {'timezones': ['Египет (UTC+02)'],
                  'alpha-2-code': 'EG',
                  'alpha-3-code': 'EGY',
                  'continent': 'Африка',
                  'name': 'Египет',
                  'capital': 'Каир'},
                 {'timezones': ['Эритрея (UTC+03)'],
                  'alpha-2-code': 'ER',
                  'alpha-3-code': 'ERI',
                  'continent': 'Африка',
                  'name': 'Эритрея',
                  'capital': 'Асмэра'},
                 {'timezones': ['Эфиопия (UTC+03)'],
                  'alpha-2-code': 'ET',
                  'alpha-3-code': 'ETH',
                  'continent': 'Африка',
                  'name': 'Эфиопия',
                  'capital': 'Аддис-Абеба'},
                 {'timezones': ['Финляндия (UTC+02)'],
                  'alpha-2-code': 'FI',
                  'alpha-3-code': 'FIN',
                  'continent': 'Европа',
                  'name': 'Финляндия',
                  'capital': 'Хельсинки'},
                 {'timezones': ['Фиджи (UTC+12)'],
                  'alpha-2-code': 'FJ',
                  'alpha-3-code': 'FJI',
                  'continent': 'Океания',
                  'name': 'Фиджи',
                  'capital': 'Сува'},
                 {'timezones': ['Франция (UTC+01)'],
                  'alpha-2-code': 'FR',
                  'alpha-3-code': 'FRA',
                  'continent': 'Европа',
                  'name': 'Франция',
                  'capital': 'Париж'},
                 {'timezones': ['Габон (UTC+01)'],
                  'alpha-2-code': 'GA',
                  'alpha-3-code': 'GAB',
                  'continent': 'Африка',
                  'name': 'Габон',
                  'capital': 'Либревиль'},
                 {'timezones': ['Грузия (UTC+04)'],
                  'alpha-2-code': 'GE',
                  'alpha-3-code': 'GEO',
                  'continent': 'Азия',
                  'name': 'Грузия',
                  'capital': 'Тбилиси'},
                 {'timezones': ['Гана (UTC)'],
                  'alpha-2-code': 'GH',
                  'alpha-3-code': 'GHA',
                  'continent': 'Африка',
                  'name': 'Гана',
                  'capital': 'Аккра'},
                 {'timezones': ['Гамбия (UTC)'],
                  'alpha-2-code': 'GM',
                  'alpha-3-code': 'GMB',
                  'continent': 'Африка',
                  'name': 'Гамбия',
                  'capital': 'Банджул'},
                 {'timezones': ['Гвинея (UTC)'],
                  'alpha-2-code': 'GN',
                  'alpha-3-code': 'GIN',
                  'continent': 'Африка',
                  'name': 'Гвинея',
                  'capital': 'Конакри'},
                 {'timezones': ['Греция (UTC+02)'],
                  'alpha-2-code': 'GR',
                  'alpha-3-code': 'GRC',
                  'continent': 'Европа',
                  'name': 'Греция',
                  'capital': 'Афины'},
                 {'timezones': ['Гватемала (UTC-06)'],
                  'alpha-2-code': 'GT',
                  'alpha-3-code': 'GTM',
                  'continent': 'Северная Америка',
                  'name': 'Гватемала',
                  'capital': 'Гватемала'},
                 {'timezones': ['Гаити (UTC-05)'],
                  'alpha-2-code': 'HT',
                  'alpha-3-code': 'HTI',
                  'continent': 'Северная Америка',
                  'name': 'Гаити',
                  'capital': 'Порт-о-Пренс'},
                 {'timezones': ['Гвинея-Бисау (UTC)'],
                  'alpha-2-code': 'GW',
                  'alpha-3-code': 'GNB',
                  'continent': 'Африка',
                  'name': 'Гвинея-Бисау',
                  'capital': 'Бисау'},
                 {'timezones': ['Гайана (UTC-04)'],
                  'alpha-2-code': 'GY',
                  'alpha-3-code': 'GUY',
                  'continent': 'Южная Америка',
                  'name': 'Гайана',
                  'capital': 'Джорджтаун'},
                 {'timezones': ['Гондурас (UTC-06)'],
                  'alpha-2-code': 'HN',
                  'alpha-3-code': 'HND',
                  'continent': 'Северная Америка',
                  'name': 'Гондурас',
                  'capital': 'Тегусигальпа'},
                 {'timezones': ['Венгрия (UTC+01)'],
                  'alpha-2-code': 'HU',
                  'alpha-3-code': 'HUN',
                  'continent': 'Европа',
                  'name': 'Венгрия',
                  'capital': 'Будапешт'},
                 {'timezones': ['Индонезия (UTC+07)',
                                'Индонезия (UTC+08)',
                                'Индонезия (UTC+09)'],
                  'alpha-2-code': 'ID',
                  'alpha-3-code': 'IDN',
                  'continent': 'Азия',
                  'name': 'Индонезия',
                  'capital': 'Джакарта'},
                 {'timezones': ['Ирландия (UTC)'],
                  'alpha-2-code': 'IE',
                  'alpha-3-code': 'IRL',
                  'continent': 'Европа',
                  'name': 'Ирландия',
                  'capital': 'Дублин'},
                 {'timezones': ['Израиль (UTC+02)'],
                  'alpha-2-code': 'IL',
                  'alpha-3-code': 'ISR',
                  'continent': 'Азия',
                  'name': 'Израиль',
                  'capital': 'Иерусалим'},
                 {'timezones': ['Индия (UTC+05:30'],
                  'alpha-2-code': 'IN',
                  'alpha-3-code': 'IND',
                  'continent': 'Азия',
                  'name': 'Индия',
                  'capital': 'Дели'},
                 {'timezones': ['Ирак (UTC+03)'],
                  'alpha-2-code': 'IQ',
                  'alpha-3-code': 'IRQ',
                  'continent': 'Азия',
                  'name': 'Ирак',
                  'capital': 'Багдад'},
                 {'timezones': ['Иран (UTC+03:30)'],
                  'alpha-2-code': 'IR',
                  'alpha-3-code': 'IRN',
                  'continent': 'Азия',
                  'name': 'Иран',
                  'capital': 'Тегеран'},
                 {'timezones': ['Исландия (UTC)'],
                  'alpha-2-code': 'IS',
                  'alpha-3-code': 'ISL',
                  'continent': 'Европа',
                  'name': 'Исландия',
                  'capital': 'Рейкьявик'},
                 {'timezones': ['Италия (UTC+01)'],
                  'alpha-2-code': 'IT',
                  'alpha-3-code': 'ITA',
                  'continent': 'Европа',
                  'name': 'Италия',
                  'capital': 'Рим'},
                 {'timezones': ['Ямайка (UTC-05)'],
                  'alpha-2-code': 'JM',
                  'alpha-3-code': 'JAM',
                  'continent': 'Северная Америка',
                  'name': 'Ямайка',
                  'capital': 'Кингстон'},
                 {'timezones': ['Иордания (UTC+02)'],
                  'alpha-2-code': 'JO',
                  'alpha-3-code': 'JOR',
                  'continent': 'Азия',
                  'name': 'Иордания',
                  'capital': 'Амман'},
                 {'timezones': ['Япония (UTC+09)'],
                  'alpha-2-code': 'JP',
                  'alpha-3-code': 'JPN',
                  'continent': 'Азия',
                  'name': 'Япония',
                  'capital': 'Токио'},
                 {'timezones': ['Кения (UTC+03)'],
                  'alpha-2-code': 'KE',
                  'alpha-3-code': 'KEN',
                  'continent': 'Африка',
                  'name': 'Кения',
                  'capital': 'Найроби'},
                 {'timezones': ['Киргизия (UTC+06)'],
                  'alpha-2-code': 'KG',
                  'alpha-3-code': 'KGZ',
                  'continent': 'Азия',
                  'name': 'Киргизия',
                  'capital': 'Бишкек'},
                 {'timezones': ['Кирибати (UTC+12)',
                                'Кирибати (UTC+13)',
                                'Кирибати (UTC+14)'],
                  'alpha-2-code': 'KI',
                  'alpha-3-code': 'KIR',
                  'continent': 'Океания',
                  'name': 'Кирибати',
                  'capital': 'Южная Тарава'},
                 {'timezones': ['КНДР (UTC+09)'],
                  'alpha-2-code': 'KP',
                  'alpha-3-code': 'PRK',
                  'continent': 'Азия',
                  'name': 'КНДР',
                  'capital': 'Пхеньян'},
                 {'timezones': ['Республика Корея (UTC+09)'],
                  'alpha-2-code': 'KR',
                  'alpha-3-code': 'KOR',
                  'continent': 'Азия',
                  'name': 'Республика Корея',
                  'capital': 'Сеул'},
                 {'timezones': ['Кувейт (UTC+03)'],
                  'alpha-2-code': 'KW',
                  'alpha-3-code': 'KWT',
                  'continent': 'Азия',
                  'name': 'Кувейт',
                  'capital': 'Эль-Кувейт'},
                 {'timezones': ['Ливан (UTC+02)'],
                  'alpha-2-code': 'LB',
                  'alpha-3-code': 'LBN',
                  'continent': 'Азия',
                  'name': 'Ливан',
                  'capital': 'Бейрут'},
                 {'timezones': ['Лихтенштейн (UTC+01)'],
                  'alpha-2-code': 'LI',
                  'alpha-3-code': 'LIE',
                  'continent': 'Европа',
                  'name': 'Лихтенштейн',
                  'capital': 'Вадуц'},
                 {'timezones': ['Либерия (UTC)'],
                  'alpha-2-code': 'LR',
                  'alpha-3-code': 'LBR',
                  'continent': 'Африка',
                  'name': 'Либерия',
                  'capital': 'Монровия'},
                 {'timezones': ['Лесото (UTC+02)'],
                  'alpha-2-code': 'LS',
                  'alpha-3-code': 'LSO',
                  'continent': 'Африка',
                  'name': 'Лесото',
                  'capital': 'Масеру'},
                 {'timezones': ['Литва (UTC+02)'],
                  'alpha-2-code': 'LT',
                  'alpha-3-code': 'LTU',
                  'continent': 'Европа',
                  'name': 'Литва',
                  'capital': 'Вильнюс'},
                 {'timezones': ['Люксембург (UTC+01)'],
                  'alpha-2-code': 'LU',
                  'alpha-3-code': 'LUX',
                  'continent': 'Европа',
                  'name': 'Люксембург',
                  'capital': 'Люксембург'},
                 {'timezones': ['Латвия (UTC+02)'],
                  'alpha-2-code': 'LV',
                  'alpha-3-code': 'LVA',
                  'continent': 'Европа',
                  'name': 'Латвия',
                  'capital': 'Рига'},
                 {'timezones': ['Ливия (UTC+02)'],
                  'alpha-2-code': 'LY',
                  'alpha-3-code': 'LBY',
                  'continent': 'Африка',
                  'name': 'Ливия',
                  'capital': 'Триполи'},
                 {'timezones': ['Мадагаскар (UTC+03)'],
                  'alpha-2-code': 'MG',
                  'alpha-3-code': 'MDG',
                  'continent': 'Африка',
                  'name': 'Мадагаскар',
                  'capital': 'Антананариву'},
                 {'timezones': ['Маршалловы Острова (UTC+12)'],
                  'alpha-2-code': 'MH',
                  'alpha-3-code': 'MHL',
                  'continent': 'Океания',
                  'name': 'Маршалловы Острова',
                  'capital': 'Маджуро'},
                 {'timezones': ['Северная Македония (UTC+01)'],
                  'alpha-2-code': 'MK',
                  'alpha-3-code': 'MKD',
                  'continent': 'Европа',
                  'name': 'Северная Македония',
                  'capital': 'Скопье'},
                 {'timezones': ['Мали (UTC)'],
                  'alpha-2-code': 'ML',
                  'alpha-3-code': 'MLI',
                  'continent': 'Африка',
                  'name': 'Мали',
                  'capital': 'Бамако'},
                 {'timezones': ['Мьянма (UTC+06:30)'],
                  'alpha-2-code': 'MM',
                  'alpha-3-code': 'MMR',
                  'continent': 'Азия',
                  'name': 'Мьянма',
                  'capital': 'Нейпьидо'},
                 {'timezones': ['Монголия (UTC+07)',
                                'Монголия (UTC+08)'],
                  'alpha-2-code': 'MN',
                  'alpha-3-code': 'MNG',
                  'continent': 'Азия',
                  'name': 'Монголия',
                  'capital': 'Улан-Батор'},
                 {'timezones': ['Мавритания (UTC)'],
                  'alpha-2-code': 'MR',
                  'alpha-3-code': 'MRT',
                  'continent': 'Африка',
                  'name': 'Мавритания',
                  'capital': 'Нуакшот'},
                 {'timezones': ['Мальта (UTC+01)'],
                  'alpha-2-code': 'MT',
                  'alpha-3-code': 'MLT',
                  'continent': 'Европа',
                  'name': 'Мальта',
                  'capital': 'Валлетта'},
                 {'timezones': ['Маврикий (UTC+04)'],
                  'alpha-2-code': 'MU',
                  'alpha-3-code': 'MUS',
                  'continent': 'Африка',
                  'name': 'Маврикий',
                  'capital': 'Порт-Луи'},
                 {'timezones': ['Мальдивы (UTC+05)'],
                  'alpha-2-code': 'MV',
                  'alpha-3-code': 'MDV',
                  'continent': 'Азия',
                  'name': 'Мальдивы',
                  'capital': 'Мале'},
                 {'timezones': ['Малави (UTC+02)'],
                  'alpha-2-code': 'MW',
                  'alpha-3-code': 'MWI',
                  'continent': 'Африка',
                  'name': 'Малави',
                  'capital': 'Лилонгве'},
                 {'timezones': ['Мексика (UTC-08)',
                                'Мексика (UTC-07)',
                                'Мексика (UTC-06)'],
                  'alpha-2-code': 'MX',
                  'alpha-3-code': 'MEX',
                  'continent': 'Северная Америка',
                  'name': 'Мексика',
                  'capital': 'Мехико'},
                 {'timezones': ['Малайзия (UTC+08)'],
                  'alpha-2-code': 'MY',
                  'alpha-3-code': 'MYS',
                  'continent': 'Азия',
                  'name': 'Малайзия',
                  'capital': 'Куала-Лумпур'},
                 {'timezones': ['Мозамбик (UTC+02)'],
                  'alpha-2-code': 'MZ',
                  'alpha-3-code': 'MOZ',
                  'continent': 'Африка',
                  'name': 'Мозамбик',
                  'capital': 'Мапуту'},
                 {'timezones': ['Намибия (UTC+01)'],
                  'alpha-2-code': 'NA',
                  'alpha-3-code': 'NAM',
                  'continent': 'Африка',
                  'name': 'Намибия',
                  'capital': 'Виндхук'},
                 {'timezones': ['Нигер (UTC+01)'],
                  'alpha-2-code': 'NE',
                  'alpha-3-code': 'NER',
                  'continent': 'Африка',
                  'name': 'Нигер',
                  'capital': 'Ниамей'},
                 {'timezones': ['Нигерия (UTC+01)'],
                  'alpha-2-code': 'NG',
                  'alpha-3-code': 'NGA',
                  'continent': 'Африка',
                  'name': 'Нигерия',
                  'capital': 'Абуджа'},
                 {'timezones': ['Никарагуа (UTC-06)'],
                  'alpha-2-code': 'NI',
                  'alpha-3-code': 'NIC',
                  'continent': 'Северная Америка',
                  'name': 'Никарагуа',
                  'capital': 'Манагуа'},
                 {'timezones': ['Нидерланды (UTC+01)'],
                  'alpha-2-code': 'NL',
                  'alpha-3-code': 'NLD',
                  'continent': 'Европа',
                  'name': 'Нидерланды',
                  'capital': 'Амстердам'},
                 {'timezones': ['Норвегия (UTC+01)'],
                  'alpha-2-code': 'NO',
                  'alpha-3-code': 'NOR',
                  'continent': 'Европа',
                  'name': 'Норвегия',
                  'capital': 'Осло'},
                 {'timezones': ['Непал (UTC+05:45'],
                  'alpha-2-code': 'NP',
                  'alpha-3-code': 'NPL',
                  'continent': 'Азия',
                  'name': 'Непал',
                  'capital': 'Катманду'},
                 {'timezones': ['Науру (UTC+12)'],
                  'alpha-2-code': 'NR',
                  'alpha-3-code': 'NRU',
                  'continent': 'Океания',
                  'name': 'Науру',
                  'capital': 'Ярен'},
                 {'timezones': ['Новая Зеландия (UTC+12)'],
                  'alpha-2-code': 'NZ',
                  'alpha-3-code': 'NZL',
                  'continent': 'Океания',
                  'name': 'Новая Зеландия',
                  'capital': 'Веллингтон'},
                 {'timezones': ['Оман (UTC+04'],
                  'alpha-2-code': 'OM',
                  'alpha-3-code': 'OMN',
                  'continent': 'Азия',
                  'name': 'Оман',
                  'capital': 'Маскат'},
                 {'timezones': ['Панама (UTC-05)'],
                  'alpha-2-code': 'PA',
                  'alpha-3-code': 'PAN',
                  'continent': 'Северная Америка',
                  'name': 'Панама',
                  'capital': 'Панама'},
                 {'timezones': ['Перу (UTC-05)'],
                  'alpha-2-code': 'PE',
                  'alpha-3-code': 'PER',
                  'continent': 'Южная Америка',
                  'name': 'Перу',
                  'capital': 'Лима'},
                 {'timezones': ['Папуа - Новая Гвинея (UTC+10)'],
                  'alpha-2-code': 'PG',
                  'alpha-3-code': 'PNG',
                  'continent': 'Океания',
                  'name': 'Папуа - Новая Гвинея',
                  'capital': 'Порт-Морсби'},
                 {'timezones': ['Филиппины (UTC+08)'],
                  'alpha-2-code': 'PH',
                  'alpha-3-code': 'PHL',
                  'continent': 'Азия',
                  'name': 'Филиппины',
                  'capital': 'Манила'},
                 {'timezones': ['Пакистан (UTC+05)'],
                  'alpha-2-code': 'PK',
                  'alpha-3-code': 'PAK',
                  'continent': 'Азия',
                  'name': 'Пакистан',
                  'capital': 'Исламабад'},
                 {'timezones': ['Польша (UTC+01)'],
                  'alpha-2-code': 'PL',
                  'alpha-3-code': 'POL',
                  'continent': 'Европа',
                  'name': 'Польша',
                  'capital': 'Варшава'},
                 {'timezones': ['Португалия (UTC)'],
                  'alpha-2-code': 'PT',
                  'alpha-3-code': 'PRT',
                  'continent': 'Европа',
                  'name': 'Португалия',
                  'capital': 'Лиссабон'},
                 {'timezones': ['Палау (UTC+09)'],
                  'alpha-2-code': 'PW',
                  'alpha-3-code': 'PLW',
                  'continent': 'Океания',
                  'name': 'Палау',
                  'capital': 'Кампала'},
                 {'timezones': ['Парагвай (UTC-04)'],
                  'alpha-2-code': 'PY',
                  'alpha-3-code': 'PRY',
                  'continent': 'Южная Америка',
                  'name': 'Парагвай',
                  'capital': 'Асунсьон'},
                 {'timezones': ['Катар (UTC+03)'],
                  'alpha-2-code': 'QA',
                  'alpha-3-code': 'QAT',
                  'continent': 'Азия',
                  'name': 'Катар',
                  'capital': 'Доха'},
                 {'timezones': ['Румыния (UTC+02)'],
                  'alpha-2-code': 'RO',
                  'alpha-3-code': 'ROU',
                  'continent': 'Европа',
                  'name': 'Румыния',
                  'capital': 'Бухарест'},
                 {'timezones': ['Россия (UTC+02)',
                                'Россия (UTC+03)',
                                'Россия (UTC+04)',
                                'Россия (UTC+05)',
                                'Россия (UTC+06)',
                                'Россия (UTC+07)',
                                'Россия (UTC+08)',
                                'Россия (UTC+09)',
                                'Россия (UTC+10)',
                                'Россия (UTC+11)',
                                'Россия (UTC+12)'],
                  'alpha-2-code': 'RU',
                  'alpha-3-code': 'RUS',
                  'continent': 'Европа',
                  'name': 'Россия',
                  'capital': 'Москва'},
                 {'timezones': ['Руанда (UTC+02)'],
                  'alpha-2-code': 'RW',
                  'alpha-3-code': 'RWA',
                  'continent': 'Африка',
                  'name': 'Руанда',
                  'capital': 'Кигали'},
                 {'timezones': ['Саудовская Аравия (UTC+03)'],
                  'alpha-2-code': 'SA',
                  'alpha-3-code': 'SAU',
                  'continent': 'Азия',
                  'name': 'Саудовская Аравия',
                  'capital': 'Эр-Рияд'},
                 {'timezones': ['Соломоновы Острова (UTC+11)'],
                  'alpha-2-code': 'SB',
                  'alpha-3-code': 'SLB',
                  'continent': 'Океания',
                  'name': 'Соломоновы Острова',
                  'capital': 'Хониара'},
                 {'timezones': ['Сейшельские острова (UTC+04)'],
                  'alpha-2-code': 'SC',
                  'alpha-3-code': 'SYC',
                  'continent': 'Африка',
                  'name': 'Сейшельские острова',
                  'capital': 'Виктория'},
                 {'timezones': ['Судан (UTC+03)'],
                  'alpha-2-code': 'SD',
                  'alpha-3-code': 'SDN',
                  'continent': 'Африка',
                  'name': 'Судан',
                  'capital': 'Хартум'},
                 {'timezones': ['Швеция (UTC+01)'],
                  'alpha-2-code': 'SE',
                  'alpha-3-code': 'SWE',
                  'continent': 'Европа',
                  'name': 'Швеци',
                  'capital': 'Стокгольм'},
                 {'timezones': ['Сингапур (UTC+08)'],
                  'alpha-2-code': 'SG',
                  'alpha-3-code': 'SGP',
                  'continent': 'Азия',
                  'name': 'Сингапур',
                  'capital': 'Сингапур'},
                 {'timezones': ['Словения (UTC+01)'],
                  'alpha-2-code': 'SI',
                  'alpha-3-code': 'SVN',
                  'continent': 'Европа',
                  'name': 'Словения',
                  'capital': 'Любляна'},
                 {'timezones': ['Словакия (UTC+01)'],
                  'alpha-2-code': 'SK',
                  'alpha-3-code': 'SVK',
                  'continent': 'Европа',
                  'name': 'Словакия',
                  'capital': 'Братислава'},
                 {'timezones': ['Сьерра-Леоне (UTC)'],
                  'alpha-2-code': 'SL',
                  'alpha-3-code': 'SLE',
                  'continent': 'Африка',
                  'name': 'Сьерра Леоне',
                  'capital': 'Фритаун'},
                 {'timezones': ['Сан-Марино (UTC+01)'],
                  'alpha-2-code': 'SM',
                  'alpha-3-code': 'SMR',
                  'continent': 'Европа',
                  'name': 'Сан-Марино',
                  'capital': 'Сан-Марино'},
                 {'timezones': ['Сенегал (UTC)'],
                  'alpha-2-code': 'SN',
                  'alpha-3-code': 'SEN',
                  'continent': 'Африка',
                  'name': 'Сенегал',
                  'capital': 'Дакар'},
                 {'timezones': ['Сомали (UTC+03)'],
                  'alpha-2-code': 'SO',
                  'alpha-3-code': 'SOM',
                  'continent': 'Африка',
                  'name': 'Сомали',
                  'capital': 'Могадишо'},
                 {'timezones': ['Суринам (UTC-03)'],
                  'alpha-2-code': 'SR',
                  'alpha-3-code': 'SUR',
                  'continent': 'Южная Америка',
                  'name': 'Суринам',
                  'capital': 'Парамарибо'},
                 {'timezones': ['Сан-Томе и Принсипи (UTC)'],
                  'alpha-2-code': 'ST',
                  'alpha-3-code': 'STP',
                  'continent': 'Африка',
                  'name': 'Сан-Томе и Принсипи',
                  'capital': 'Сан-Томе'},
                 {'timezones': ['Сирия (UTC+02)'],
                  'alpha-2-code': 'SY',
                  'alpha-3-code': 'SYR',
                  'continent': 'Азия',
                  'name': 'Сирия',
                  'capital': 'Дамаск'},
                 {'timezones': ['Того (UTC)'],
                  'alpha-2-code': 'TG',
                  'alpha-3-code': 'TGO',
                  'continent': 'Африка',
                  'name': 'Того',
                  'capital': 'Ломе'},
                 {'timezones': ['Таиланд (UTC+07)'],
                  'alpha-2-code': 'TH',
                  'alpha-3-code': 'THA',
                  'continent': 'Азия',
                  'name': 'Таиланд',
                  'capital': 'Бангкок'},
                 {'timezones': ['Таджикистан (UTC+05)'],
                  'alpha-2-code': 'TJ',
                  'alpha-3-code': 'TJK',
                  'continent': 'Азия',
                  'name': 'Таджикистан',
                  'capital': 'Душанбе'},
                 {'timezones': ['Туркмения (UTC+05)'],
                  'alpha-2-code': 'TM',
                  'alpha-3-code': 'TKM',
                  'continent': 'Азия',
                  'name': 'Туркмения',
                  'capital': 'Ашхабад'},
                 {'timezones': ['Тунис (UTC+01)'],
                  'alpha-2-code': 'TN',
                  'alpha-3-code': 'TUN',
                  'continent': 'Африка',
                  'name': 'Тунис',
                  'capital': 'Тунис'},
                 {'timezones': ['Тонга (UTC+13)'],
                  'alpha-2-code': 'TO',
                  'alpha-3-code': 'TON',
                  'continent': 'Океания',
                  'name': 'Тонга',
                  'capital': 'Нукуалофа'},
                 {'timezones': ['Турция (UTC+02)'],
                  'alpha-2-code': 'TR',
                  'alpha-3-code': 'TUR',
                  'continent': 'Азия',
                  'name': 'Турция',
                  'capital': 'Анкара'},
                 {'timezones': ['Тринидад и Тобаго (UTC-04)'],
                  'alpha-2-code': 'TT',
                  'alpha-3-code': 'TTO',
                  'continent': 'Северная Америка',
                  'name': 'Тринидад и Тобаго',
                  'capital': 'Порт-оф-Спейн'},
                 {'timezones': ['Тувалу (UTC+12)'],
                  'alpha-2-code': 'TV',
                  'alpha-3-code': 'TUV',
                  'continent': 'Океания',
                  'name': 'Тувалу',
                  'capital': 'Фунафути'},
                 {'timezones': ['Танзания (UTC+03)'],
                  'alpha-2-code': 'TZ',
                  'alpha-3-code': 'TZA',
                  'continent': 'Африка',
                  'name': 'Танзания',
                  'capital': 'Додома'},
                 {'timezones': ['Украина (UTC+02)',
                                'Украина (UTC+03)'],
                  'alpha-2-code': 'UA',
                  'alpha-3-code': 'UKR',
                  'continent': 'Европа',
                  'name': 'Украина',
                  'capital': 'Киев'},
                 {'timezones': ['Уганда (UTC+03)'],
                  'alpha-2-code': 'UG',
                  'alpha-3-code': 'UGA',
                  'continent': 'Африка',
                  'name': 'Уганда',
                  'capital': 'Кампала'},
                 {'timezones': ['США (UTC-11)',
                                'США (UTC-10)',
                                'США (UTC-09)',
                                'США (UTC-08)',
                                'США (UTC-07)',
                                'США (UTC-06)',
                                'США (UTC-05)',
                                'США (UTC-04)',
                                'США (UTC+10)'],
                  'alpha-2-code': 'US',
                  'alpha-3-code': 'USA',
                  'continent': 'Северная Америка',
                  'name': 'США',
                  'capital': 'Вашингтон'},
                 {'timezones': ['Уругвай (UTC-03)'],
                  'alpha-2-code': 'UY',
                  'alpha-3-code': 'URY',
                  'continent': 'Южная Америка',
                  'name': 'Уругвай',
                  'capital': 'Монтевидео'},
                 {'timezones': ['Узбекистан (UTC+05)'],
                  'alpha-2-code': 'UZ',
                  'alpha-3-code': 'UZB',
                  'continent': 'Азия',
                  'name': 'Узбекистан',
                  'capital': 'Ташкент'},
                 {'timezones': ['Ватикан (UTC+01)'],
                  'alpha-2-code': 'VA',
                  'alpha-3-code': 'VAT',
                  'continent': 'Европа',
                  'name': 'Ватикан',
                  'capital': 'Ватикан'},
                 {'timezones': ['Венесуэла (UTC-04:30)'],
                  'alpha-2-code': 'VE',
                  'alpha-3-code': 'VEN',
                  'continent': 'Южная Америка',
                  'name': 'Венесуэла',
                  'capital': 'Каракас'},
                 {'timezones': ['Вьетнам (UTC+07)'],
                  'alpha-2-code': 'VN',
                  'alpha-3-code': 'VNM',
                  'continent': 'Азия',
                  'name': 'Вьетнам',
                  'capital': 'Ханой'},
                 {'timezones': ['Вануату (UTC+11)'],
                  'alpha-2-code': 'VU',
                  'alpha-3-code': 'VUT',
                  'continent': 'Океания',
                  'name': 'Вануату',
                  'capital': 'Порт-Вила'},
                 {'timezones': ['Йемен (UTC+03)'],
                  'alpha-2-code': 'YE',
                  'alpha-3-code': 'YEM',
                  'continent': 'Азия',
                  'name': 'Йемен',
                  'capital': "Сана"},
                 {'timezones': ['Замбия (UTC+02)'],
                  'alpha-2-code': 'ZM',
                  'alpha-3-code': 'ZMB',
                  'continent': 'Африка',
                  'name': 'Замбия',
                  'capital': 'Лусака'},
                 {'timezones': ['Зимбабве (UTC+02)'],
                  'alpha-2-code': 'ZW',
                  'alpha-3-code': 'ZWE',
                  'continent': 'Африка',
                  'name': 'Зимбабве',
                  'capital': 'Хараре'},
                 {'timezones': ['Алжир (UTC+01)'],
                  'alpha-2-code': 'DZ',
                  'alpha-3-code': 'DZA',
                  'continent': 'Африка',
                  'name': 'Алжир',
                  'capital': 'Алжир'},
                 {'timezones': ['Босния и Герцеговина (UTC+01)'],
                  'alpha-2-code': 'BA',
                  'alpha-3-code': 'BIH',
                  'continent': 'Европа',
                  'name': 'Босния и Герцеговина',
                  'capital': 'Сараево'},
                 {'timezones': ['Камбоджа (UTC+07)'],
                  'alpha-2-code': 'KH',
                  'alpha-3-code': 'KHM',
                  'continent': 'Азия',
                  'name': 'Камбоджа',
                  'capital': 'Пномпень'},
                 {'timezones': ['ЦАР (UTC+01)'],
                  'alpha-2-code': 'CF',
                  'alpha-3-code': 'CAF',
                  'continent': 'Африка',
                  'name': 'ЦАР',
                  'capital': 'Банги'},
                 {'timezones': ['Чад (UTC+01)'],
                  'alpha-2-code': 'TD',
                  'alpha-3-code': 'TCD',
                  'continent': 'Африка',
                  'name': 'Чад',
                  'capital': "Нджамена"},
                 {'timezones': ['Коморы (UTC+03)'],
                  'alpha-2-code': 'KM',
                  'alpha-3-code': 'COM',
                  'continent': 'Африка',
                  'name': 'Коморы',
                  'capital': 'Морони'},
                 {'timezones': ['Хорватия (UTC+01)'],
                  'alpha-2-code': 'HR',
                  'alpha-3-code': 'HRV',
                  'continent': 'Европа',
                  'name': 'Хорватия',
                  'capital': 'Загреб'},
                 {'timezones': ['Восточный Тимор (UTC+09)'],
                  'alpha-2-code': 'TL',
                  'alpha-3-code': 'TLS',
                  'continent': 'Азия',
                  'name': 'Восточный Тимор',
                  'capital': 'Дили'},
                 {'timezones': ['Сальвадор (UTC-06)'],
                  'alpha-2-code': 'SV',
                  'alpha-3-code': 'SLV',
                  'continent': 'Северная Америка',
                  'name': 'Сальвадор',
                  'capital': 'Сан-Сальвадор'},
                 {'timezones': ['Экваториальная Гвинея (UTC+01)'],
                  'alpha-2-code': 'GQ',
                  'alpha-3-code': 'GNQ',
                  'continent': 'Африка',
                  'name': 'Экваториальная Гвинея',
                  'capital': 'Малабо'},
                 {'timezones': ['Гренада (UTC-04)'],
                  'alpha-2-code': 'GD',
                  'alpha-3-code': 'GRD',
                  'continent': 'Северная Америка',
                  'name': 'Гренада',
                  'capital': "Сент-Джорджес"},
                 {'timezones': ['Казахстан (UTC+05)',
                                'Казахстан (UTC+06)'],
                  'alpha-2-code': 'KZ',
                  'alpha-3-code': 'KAZ',
                  'continent': 'Азия',
                  'name': 'Казахстан',
                  'capital': 'Нур-Султан (Астана)'},
                 {'timezones': ['Лаос (UTC+07)'],
                  'alpha-2-code': 'LA',
                  'alpha-3-code': 'LAO',
                  'continent': 'Азия',
                  'name': 'Лаос',
                  'capital': 'Вьентьян'},
                 {'timezones': ['Микронезия (UTC+10)',
                                'Микронезия (UTC+11)'],
                  'alpha-2-code': 'FM',
                  'alpha-3-code': 'FSM',
                  'continent': 'Океания',
                  'name': 'Микронезия',
                  'capital': 'Паликир'},
                 {'timezones': ['Молдавия (UTC+02)'],
                  'alpha-2-code': 'MD',
                  'alpha-3-code': 'MDA',
                  'continent': 'Европа',
                  'name': 'Молдавия',
                  'capital': 'Кишинев'},
                 {'timezones': ['Монако (UTC+01)'],
                  'alpha-2-code': 'MC',
                  'alpha-3-code': 'MCO',
                  'continent': 'Европа',
                  'name': 'Монако',
                  'capital': 'Монако'},
                 {'timezones': ['Черногория (UTC+01)'],
                  'alpha-2-code': 'ME',
                  'alpha-3-code': 'MNE',
                  'continent': 'Европа',
                  'name': 'Черногория',
                  'capital': 'Подгорица'},
                 {'timezones': ['Марокко (UTC)'],
                  'alpha-2-code': 'MA',
                  'alpha-3-code': 'MAR',
                  'continent': 'Африка',
                  'name': 'Марокко',
                  'capital': 'Рабат'},
                 {'timezones': ['Сент-Китс и Невис (UTC-04)'],
                  'alpha-2-code': 'KN',
                  'alpha-3-code': 'KNA',
                  'continent': 'Северная Америка',
                  'name': 'Сент-Китс и Невис',
                  'capital': 'Бастер'},
                 {'timezones': ['Сент-Люсия (UTC-04)'],
                  'alpha-2-code': 'LC',
                  'alpha-3-code': 'LCA',
                  'continent': 'Северная Америка',
                  'name': 'Сент-Люсия',
                  'capital': 'Кастри'},
                 {'timezones': ['Сент-Винсент и Гренадины (UTC-04)'],
                  'alpha-2-code': 'VC',
                  'alpha-3-code': 'VCT',
                  'continent': 'Северная Америка',
                  'name': 'Сент-Винсент и Гренадины',
                  'capital': 'Кингстаун'},
                 {'timezones': ['Самоа (UTC+13)'],
                  'alpha-2-code': 'WS',
                  'alpha-3-code': 'WSM',
                  'continent': 'Океания',
                  'name': 'Самоа',
                  'capital': 'Апиа'},
                 {'timezones': ['Сербия (UTC+01)'],
                  'alpha-2-code': 'RS',
                  'alpha-3-code': 'SRB',
                  'continent': 'Европа',
                  'name': 'Сербия',
                  'capital': 'Белград'},
                 {'timezones': ['ЮАР (UTC+02)'],
                  'alpha-2-code': 'ZA',
                  'alpha-3-code': 'ZAF',
                  'continent': 'Африка',
                  'name': 'ЮАР',
                  'capital': 'Претория'},
                 {'timezones': ['Испания (UTC)',
                                'Испания (UTC+01)'],
                  'alpha-2-code': 'ES',
                  'alpha-3-code': 'ESP',
                  'continent': 'Европа',
                  'name': 'Испания',
                  'capital': 'Мадрид'},
                 {'timezones': ['Шри-Ланка (UTC+05:30)'],
                  'alpha-2-code': 'LK',
                  'alpha-3-code': 'LKA',
                  'continent': 'Азия',
                  'name': 'Шри-Ланка',
                  'capital': 'Шри-Джаяварденепура-Котте'},
                 {'timezones': ['Эсватини (Свазиленд) (UTC+02)'],
                  'alpha-2-code': 'SZ',
                  'alpha-3-code': 'SWZ',
                  'continent': 'Африка',
                  'name': 'Эсватини (Свазиленд)',
                  'capital': 'Мбабане'},
                 {'timezones': ['Швейцария (UTC+01)'],
                  'alpha-2-code': 'CH',
                  'alpha-3-code': 'CHE',
                  'continent': 'Европа',
                  'name': 'Швейцария',
                  'capital': 'Берн'},
                 {'timezones': ['ОАЭ (UTC+04)'],
                  'alpha-2-code': 'AE',
                  'alpha-3-code': 'ARE',
                  'continent': 'Азия',
                  'name': 'ОАЭ',
                  'capital': 'Абу-Даби'},
                 {'timezones': ['Великобритания (UTC)'],
                  'alpha-2-code': 'GB',
                  'alpha-3-code': 'GBR',
                  'continent': 'Европа',
                  'name': 'Великобритания',
                  'capital': 'Лондон'},
                 ]

    def day_of_week(self):
        day = self.date('%w')
        return self.DAY_NAMES[day]

    def month_name(self):
        month = self.month()
        return self.MONTH_NAMES[month]



from .. import Provider as BaseProvider


# Ukrainian job names taken from
# https://uk.wikipedia.org/wiki/%D0%A1%D0%BF%D0%B8%D1%81%D0%BE%D0%BA_%D0%BF%D1%80%D0%BE%D1%84%D0%B5%D1%81%D1%96%D0%B9
# on 22th September 2014
class Provider(BaseProvider):
    jobs = [
        # А
        'Агроном', 'Адвокат', 'Актор', 'Акушер', 'Антрополог', 'Архітектор', 'Археолог', 'Астронавт', 'Астроном',
        'Астрофізик',
        # Б
        'Бібліограф', 'Біолог', 'Бізнесмен', 'Ботанік', 'Будник', 'Бухгалтер', 'Бібліотекар',
        # В
        'Ветеринар', 'Випробувач', 'Водій', 'Вчитель', 'Візажист',
        # Г
        'Географ', 'Геолог', 'Геофізик', 'Гицель', 'Гінеколог', 'Гірник', 'Гірничий інженер', 'Головний меркшейдер',
        'Графік', 'Громадський діяч',
        # Ґ
        'Ґрунтознавець',
        # Д
        'Державний службовець', 'Дерун', 'Детектив', 'Дизайнер', 'Дипломат', 'Диригент', 'Доцент', 'Драматург',
        'Ді-джей', 'Дантист',
        # Е
        'Економіст', 'Електрик', 'Електромонтер', 'Електромонтажник', 'Електрослюсар', 'Електротехнік', 'Епідеміолог',
        'Етнограф',
        # Є
        'Євнух', 'Єгер',
        # Ж
        'Журналіст', 'Живописець',
        # З
        'Золотар', 'Зоолог',
        # І
        'Інженер', 'Історик',
        # К
        'Каскадер', 'Кінорежисер', 'Клавішник', 'Клоун', 'Композитор', 'Конструктор', 'Краєзнавець', 'Кушнір',
        'Кіноактор', 'Кінокритик', 'Кінорежисер', 'Кур\'єр', 'Кухар', 'Кінолог', 'Круп\'є',
        # Л
        'Льотчик', 'Лікар', 'Літературознавець', 'Локсмайстер',
        # М
        'Математик', 'Машиніст', 'Медик', 'Менеджер', 'Мистецтвознавець', 'Мірошник', 'Мікробіолог', 'Мінералог',
        'Мовознавець', 'Модель', 'Модельєр', 'Музикант', 'Музикознавець', 'Музичний редактор', 'Маркетолог',
        'М\'ясник',
        # Н
        'Намотувальник', 'Науковець', 'Няня', 'Нотаріус',
        # П
        'Палеонтолог', 'Паралегал', 'Парфюмер', 'Патологоанатом', 'Педагог', 'Пекар', 'Перекладач', 'Петрограф',
        'Письменник', 'Піаніст', 'Підприємець', 'Пілот', 'Правник', 'Програміст', 'Провізор', 'Прокурор',
        'Промисловець', 'Професор', 'Психолог', 'Публіцист', 'Продюсер', 'Паблік рилейшнз',
        # Р
        'Режисер', 'Різноробочий', 'Реабілітолог', 'Редактор', 'Реставратор', 'Ріелтор',
        # С
        'Сантехнік', 'Складальник', 'Скульптор', 'Соціолог', 'Співак', 'Сценарист', 'Стропальник', 'Стоматолог',
        'Слідчий', 'Стиліст', 'Секретар',
        # Ф
        'Фармацевт', 'Фермер', 'Фізик', 'Філолог', 'Фольклорист', 'Фотограф', 'Фрілансер', 'Футболіст', 'Флорист',
        # Х
        'Хімік', 'Художник', 'Хореограф',
        # Ш
        'Шериф', 'Швачка', 'Штукатур',
        # Ю
        'Ювелір', 'Юрист',
    ]



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Imposter encodings module that installs a coverage-style tracer.

This is NOT the encodings module; it is an imposter that sets up tracing
instrumentation and then replaces itself with the real encodings module.

If the directory that holds this file is placed first in the PYTHONPATH when
using "coverage" to run Python's tests, then this file will become the very
first module imported by the internals of Python 3.  It installs a
coverage.py-compatible trace function that can watch Standard Library modules
execute from the very earliest stages of Python's own boot process.  This fixes
a problem with coverage.py - that it starts too late to trace the coverage of
many of the most fundamental modules in the Standard Library.

"""

import sys

class FullCoverageTracer(object):
    def __init__(self):
        # `traces` is a list of trace events.  Frames are tricky: the same
        # frame object is used for a whole scope, with new line numbers
        # written into it.  So in one scope, all the frame objects are the
        # same object, and will eventually all will point to the last line
        # executed.  So we keep the line numbers alongside the frames.
        # The list looks like:
        #
        #   traces = [
        #       ((frame, event, arg), lineno), ...
        #       ]
        #
        self.traces = []

    def fullcoverage_trace(self, *args):
        frame, event, arg = args
        self.traces.append((args, frame.f_lineno))
        return self.fullcoverage_trace

sys.settrace(FullCoverageTracer().fullcoverage_trace)

# In coverage/files.py is actual_filename(), which uses glob.glob.  I don't
# understand why, but that use of glob borks everything if fullcoverage is in
# effect.  So here we make an ugly hail-mary pass to switch off glob.glob over
# there.  This means when using fullcoverage, Windows path names will not be
# their actual case.

#sys.fullcoverage = True

# Finally, remove our own directory from sys.path; remove ourselves from
# sys.modules; and re-import "encodings", which will be the real package
# this time.  Note that the delete from sys.modules dictionary has to
# happen last, since all of the symbols in this module will become None
# at that exact moment, including "sys".

parentdir = max(filter(__file__.startswith, sys.path), key=len)
sys.path.remove(parentdir)
del sys.modules['encodings']
import encodings



from django.db.models import Manager
from django.conf import settings as dj_settings
from rest_framework import serializers

from .utils import get_image_urls
from . import settings
from .models import (
    Page,
    HeroModule,
    TwoColumnModule,
    IconListModule,
    IconListItem,
    HeaderModule,
    FAQModule,
    FAQItem,
    LocationModule,
    LocationItem,
    ImageGalleryModule,
    GalleryImage,
    CarouselModule,
    CarouselItem,
)


class PageListSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        iterable = data.all() if isinstance(data, Manager) else data
        return [self.child.to_representation(item, many=True) for item in iterable]


class PageSerializer(serializers.ModelSerializer):
    og_image = serializers.SerializerMethodField()

    def get_og_image(self, obj: Page) -> str:
        return get_image_urls(obj, "og_image")

    @classmethod
    def many_init(cls, *args, **kwargs):
        kwargs["child"] = cls()
        return PageListSerializer(*args, **kwargs)

    def to_representation(self, instance, many=False):
        orig_rep = super().to_representation(instance)

        request = self.context.get("request", None)
        url = (
            request.build_absolute_uri().replace("/v1/", "/")
            if request
            else dj_settings.SITE_URL
        )

        head = {
            "type": "head",
            "locale": orig_rep.get("locale", "en-US"),
            "title": orig_rep.get("title", "Wonderful"),
            "description": orig_rep.get("description", ""),
            "keywords": orig_rep.get("keywords", ""),
            "url": url,
            "gtm": orig_rep.get("gtm", "gtm-Test"),
            "open_graph": {
                "type": "website",
                "url": url,
                "locale": orig_rep.get("locale", "en-US"),
                "title": orig_rep.get("title", "Wonderful"),
                "description": orig_rep.get("seo_description", ""),
                "image": orig_rep.get("og_image", ""),
                "site_name": orig_rep.get("site_name", "Wonderful"),
                "card": "summary_large_image",
            },
        }

        new_rep = {"page": [head]}

        if many:
            body = {}
            for key in self.Meta.body_fields:
                body[key] = orig_rep.get(key)

            new_rep["page"].append(body)
        else:
            for key in self.Meta.body_fields:
                if not orig_rep.get(key):
                    continue

                new_rep["page"].append(
                    {"type": key, "items": orig_rep.get(key),}
                )

        return new_rep

    class Meta:
        model = Page
        body_fields = (
            "title",
            "slug",
            "description",
            "keywords",
            "og_image",
            "site_name",
        )
        fields = body_fields


class HeroModuleSerializer(serializers.ModelSerializer):
    image = serializers.SerializerMethodField()

    def get_image(self, obj: HeroModule) -> dict:
        images_json = get_image_urls(obj, "image", settings.WIDE_IMAGE_SIZES)
        images_json["altText"] = obj.image_alt_text
        return images_json

    class Meta:
        model = HeroModule
        fields = (
            "kind",
            "image",
            "header",
            "subheader",
            "button_url",
            "button_text",
            "style",
            "behavior",
        )


class TwoColumnModuleSerializer(serializers.ModelSerializer):
    left_image = serializers.SerializerMethodField()
    right_image = serializers.SerializerMethodField()

    def get_left_image(self, obj: TwoColumnModule) -> dict:
        images_json = get_image_urls(obj, "left_image", settings.SQUARE_IMAGE_SIZES)
        images_json["altText"] = obj.left_image_alt_text
        return images_json

    def get_right_image(self, obj: TwoColumnModule) -> dict:
        images_json = get_image_urls(obj, "right_image", settings.SQUARE_IMAGE_SIZES)
        images_json["altText"] = obj.right_image_alt_text
        return images_json

    class Meta:
        model = TwoColumnModule
        fields = (
            "kind",
            "left_image",
            "left_header",
            "left_description",
            "left_button_url",
            "left_button_text",
            "right_image",
            "right_header",
            "right_description",
            "right_button_url",
            "right_button_text",
            "style",
            "behavior",
        )


class HeaderModuleSerializer(serializers.ModelSerializer):
    class Meta:
        model = HeaderModule
        fields = (
            "header",
            "subheader",
            "style",
            "behavior",
        )


class IconListItemSerializer(serializers.ModelSerializer):
    icon = serializers.SerializerMethodField()

    def get_icon(self, obj):
        images_json = get_image_urls(obj, "icon")
        images_json["altText"] = obj.icon_alt_text
        return images_json

    class Meta:
        model = IconListItem
        field = (
            "kind",
            "icon",
            "text",
        )


class IconListModuleSerializer(serializers.ModelSerializer):
    items = IconListItemSerializer(many=True)

    class Meta:
        model = IconListModule
        fields = (
            "kind",
            "title",
            "subtitle",
            "footnote",
            "items",
            "style",
            "behavior",
        )


class FAQItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = FAQModule
        field = (
            "kind",
            "question",
            "answer",
        )


class FAQModuleSerializer(serializers.ModelSerializer):
    items = IconListItemSerializer(many=True)

    class Meta:
        model = FAQItem
        fields = (
            "kind",
            "items",
            "style",
            "behavior",
        )


class LocationItemSerializer(serializers.ModelSerializer):
    class Meta:
        model = FAQModule
        field = (
            "kind",
            "title",
            "address1",
            "address2",
            "city",
            "state",
            "zip_code",
            "country",
            "hours",
            "phone_number",
        )


class LocationModuleSerializer(serializers.ModelSerializer):
    items = IconListItemSerializer(many=True)

    class Meta:
        model = LocationItem
        fields = (
            "kind",
            "items",
            "style",
            "behavior",
        )


class GalleryImageSerializer(serializers.ModelSerializer):
    image = serializers.SerializerMethodField()

    def get_image(self, obj: GalleryImage) -> dict:
        images_json = get_image_urls(obj, "image", settings.GALLERY_IMAGE_SIZES)
        images_json["altText"] = obj.image_alt_text
        return images_json

    class Meta:
        model = GalleryImage
        field = (
            "kind",
            "image",
        )


class ImageGallerySerializer(serializers.ModelSerializer):
    items = IconListItemSerializer(many=True)

    class Meta:
        model = ImageGalleryModule
        fields = (
            "kind",
            "items",
            "style",
            "behavior",
        )


class CarouselItemSerializer(serializers.ModelSerializer):
    image = serializers.SerializerMethodField()

    def get_image(self, obj: CarouselItem) -> dict:
        images_json = get_image_urls(obj, "image", settings.CAROUSEL_IMAGE_SIZES)
        images_json["altText"] = obj.image_alt_text
        return images_json

    class Meta:
        model = CarouselItem
        field = (
            "kind",
            "title",
            "subtitle",
            "image",
            "button_url",
            "button_text",
        )


class CarouselModuleSerializer(serializers.ModelSerializer):
    items = IconListItemSerializer(many=True)

    class Meta:
        model = CarouselModule
        fields = (
            "kind",
            "button_url",
            "button_text",
            "items",
            "style",
            "behavior",
        )


## You are here. Swap out all the modules with the new ones from above.
class PageDetailsSerializer(PageSerializer):
    hero_modules = HeroModuleSerializer(
        source="page_hero_modules", read_only=True, many=True
    )
    two_column_modules = TwoColumnModuleSerializer(
        source="page_two_column_modules", read_only=True, many=True
    )
    icon_list_modules = IconListModuleSerializer(
        source="page_icon_list_modules", read_only=True, many=True
    )
    header_modules = HeaderModuleSerializer(
        source="page_header_modules", read_only=True, many=True
    )
    faq_modules = FAQModuleSerializer(
        source="page_faq_modules", read_only=True, many=True
    )
    location_modules = LocationModuleSerializer(
        source="page_location_modules", read_only=True, many=True
    )
    image_gallery_modules = ImageGallerySerializer(
        source="page_gallery_modules", read_only=True, many=True
    )
    carousel_modules = CarouselModuleSerializer(
        source="page_carousel_modules", read_only=True, many=True
    )

    class Meta:
        model = Page
        body_fields = (
            "hero_modules",
            "two_column_modules",
            "icon_list_modules",
            "header_modules",
            "faq_modules",
            "location_modules",
            "image_gallery_modules",
            "carousel_modules",
        )
        fields = PageSerializer.Meta.fields + body_fields



from collections import namedtuple

# Structure returned by DatabaseIntrospection.get_table_list()
TableInfo = namedtuple('TableInfo', ['name', 'type'])

# Structure returned by the DB-API cursor.description interface (PEP 249)
FieldInfo = namedtuple('FieldInfo', 'name type_code display_size internal_size precision scale null_ok default')


class BaseDatabaseIntrospection:
    """Encapsulate backend-specific introspection utilities."""
    data_types_reverse = {}

    def __init__(self, connection):
        self.connection = connection

    def get_field_type(self, data_type, description):
        """
        Hook for a database backend to use the cursor description to
        match a Django field type to a database column.

        For Oracle, the column data_type on its own is insufficient to
        distinguish between a FloatField and IntegerField, for example.
        """
        return self.data_types_reverse[data_type]

    def identifier_converter(self, name):
        """
        Apply a conversion to the identifier for the purposes of comparison.

        The default identifier converter is for case sensitive comparison.
        """
        return name

    def table_names(self, cursor=None, include_views=False):
        """
        Return a list of names of all tables that exist in the database.
        Sort the returned table list by Python's default sorting. Do NOT use
        the database's ORDER BY here to avoid subtle differences in sorting
        order between databases.
        """
        def get_names(cursor):
            return sorted(ti.name for ti in self.get_table_list(cursor)
                          if include_views or ti.type == 't')
        if cursor is None:
            with self.connection.cursor() as cursor:
                return get_names(cursor)
        return get_names(cursor)

    def get_table_list(self, cursor):
        """
        Return an unsorted list of TableInfo named tuples of all tables and
        views that exist in the database.
        """
        raise NotImplementedError('subclasses of BaseDatabaseIntrospection may require a get_table_list() method')

    def django_table_names(self, only_existing=False, include_views=True):
        """
        Return a list of all table names that have associated Django models and
        are in INSTALLED_APPS.

        If only_existing is True, include only the tables in the database.
        """
        from django.apps import apps
        from django.db import router
        tables = set()
        for app_config in apps.get_app_configs():
            for model in router.get_migratable_models(app_config, self.connection.alias):
                if not model._meta.managed:
                    continue
                tables.add(model._meta.db_table)
                tables.update(
                    f.m2m_db_table() for f in model._meta.local_many_to_many
                    if f.remote_field.through._meta.managed
                )
        tables = list(tables)
        if only_existing:
            existing_tables = set(self.table_names(include_views=include_views))
            tables = [
                t
                for t in tables
                if self.identifier_converter(t) in existing_tables
            ]
        return tables

    def installed_models(self, tables):
        """
        Return a set of all models represented by the provided list of table
        names.
        """
        from django.apps import apps
        from django.db import router
        all_models = []
        for app_config in apps.get_app_configs():
            all_models.extend(router.get_migratable_models(app_config, self.connection.alias))
        tables = set(map(self.identifier_converter, tables))
        return {
            m for m in all_models
            if self.identifier_converter(m._meta.db_table) in tables
        }

    def sequence_list(self):
        """
        Return a list of information about all DB sequences for all models in
        all apps.
        """
        from django.apps import apps
        from django.db import router

        sequence_list = []
        with self.connection.cursor() as cursor:
            for app_config in apps.get_app_configs():
                for model in router.get_migratable_models(app_config, self.connection.alias):
                    if not model._meta.managed:
                        continue
                    if model._meta.swapped:
                        continue
                    sequence_list.extend(self.get_sequences(cursor, model._meta.db_table, model._meta.local_fields))
                    for f in model._meta.local_many_to_many:
                        # If this is an m2m using an intermediate table,
                        # we don't need to reset the sequence.
                        if f.remote_field.through._meta.auto_created:
                            sequence = self.get_sequences(cursor, f.m2m_db_table())
                            sequence_list.extend(sequence or [{'table': f.m2m_db_table(), 'column': None}])
        return sequence_list

    def get_sequences(self, cursor, table_name, table_fields=()):
        """
        Return a list of introspected sequences for table_name. Each sequence
        is a dict: {'table': <table_name>, 'column': <column_name>}. An optional
        'name' key can be added if the backend supports named sequences.
        """
        raise NotImplementedError('subclasses of BaseDatabaseIntrospection may require a get_sequences() method')

    def get_key_columns(self, cursor, table_name):
        """
        Backends can override this to return a list of:
            (column_name, referenced_table_name, referenced_column_name)
        for all key columns in given table.
        """
        raise NotImplementedError('subclasses of BaseDatabaseIntrospection may require a get_key_columns() method')

    def get_primary_key_column(self, cursor, table_name):
        """
        Return the name of the primary key column for the given table.
        """
        for constraint in self.get_constraints(cursor, table_name).values():
            if constraint['primary_key']:
                return constraint['columns'][0]
        return None

    def get_constraints(self, cursor, table_name):
        """
        Retrieve any constraints or keys (unique, pk, fk, check, index)
        across one or more columns.

        Return a dict mapping constraint names to their attributes,
        where attributes is a dict with keys:
         * columns: List of columns this covers
         * primary_key: True if primary key, False otherwise
         * unique: True if this is a unique constraint, False otherwise
         * foreign_key: (table, column) of target, or None
         * check: True if check constraint, False otherwise
         * index: True if index, False otherwise.
         * orders: The order (ASC/DESC) defined for the columns of indexes
         * type: The type of the index (btree, hash, etc.)

        Some backends may return special constraint names that don't exist
        if they don't name constraints of a certain type (e.g. SQLite)
        """
        raise NotImplementedError('subclasses of BaseDatabaseIntrospection may require a get_constraints() method')



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "scarlet.pagebuilder"

    def ready(self):
        super().ready()
        self.module.autodiscover()



from collections import defaultdict

from django.apps import apps
from django.db import models
from django.utils.translation import gettext_lazy as _


class ContentTypeManager(models.Manager):
    use_in_migrations = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Cache shared by all the get_for_* methods to speed up
        # ContentType retrieval.
        self._cache = {}

    def get_by_natural_key(self, app_label, model):
        try:
            ct = self._cache[self.db][(app_label, model)]
        except KeyError:
            ct = self.get(app_label=app_label, model=model)
            self._add_to_cache(self.db, ct)
        return ct

    def _get_opts(self, model, for_concrete_model):
        if for_concrete_model:
            model = model._meta.concrete_model
        return model._meta

    def _get_from_cache(self, opts):
        key = (opts.app_label, opts.model_name)
        return self._cache[self.db][key]

    def get_for_model(self, model, for_concrete_model=True):
        """
        Return the ContentType object for a given model, creating the
        ContentType if necessary. Lookups are cached so that subsequent lookups
        for the same model don't hit the database.
        """
        opts = self._get_opts(model, for_concrete_model)
        try:
            return self._get_from_cache(opts)
        except KeyError:
            pass

        # The ContentType entry was not found in the cache, therefore we
        # proceed to load or create it.
        try:
            # Start with get() and not get_or_create() in order to use
            # the db_for_read (see #20401).
            ct = self.get(app_label=opts.app_label, model=opts.model_name)
        except self.model.DoesNotExist:
            # Not found in the database; we proceed to create it. This time
            # use get_or_create to take care of any race conditions.
            ct, created = self.get_or_create(
                app_label=opts.app_label,
                model=opts.model_name,
            )
        self._add_to_cache(self.db, ct)
        return ct

    def get_for_models(self, *models, for_concrete_models=True):
        """
        Given *models, return a dictionary mapping {model: content_type}.
        """
        results = {}
        # Models that aren't already in the cache.
        needed_app_labels = set()
        needed_models = set()
        # Mapping of opts to the list of models requiring it.
        needed_opts = defaultdict(list)
        for model in models:
            opts = self._get_opts(model, for_concrete_models)
            try:
                ct = self._get_from_cache(opts)
            except KeyError:
                needed_app_labels.add(opts.app_label)
                needed_models.add(opts.model_name)
                needed_opts[opts].append(model)
            else:
                results[model] = ct
        if needed_opts:
            # Lookup required content types from the DB.
            cts = self.filter(
                app_label__in=needed_app_labels,
                model__in=needed_models
            )
            for ct in cts:
                opts_models = needed_opts.pop(ct.model_class()._meta, [])
                for model in opts_models:
                    results[model] = ct
                self._add_to_cache(self.db, ct)
        # Create content types that weren't in the cache or DB.
        for opts, opts_models in needed_opts.items():
            ct = self.create(
                app_label=opts.app_label,
                model=opts.model_name,
            )
            self._add_to_cache(self.db, ct)
            for model in opts_models:
                results[model] = ct
        return results

    def get_for_id(self, id):
        """
        Lookup a ContentType by ID. Use the same shared cache as get_for_model
        (though ContentTypes are obviously not created on-the-fly by get_by_id).
        """
        try:
            ct = self._cache[self.db][id]
        except KeyError:
            # This could raise a DoesNotExist; that's correct behavior and will
            # make sure that only correct ctypes get stored in the cache dict.
            ct = self.get(pk=id)
            self._add_to_cache(self.db, ct)
        return ct

    def clear_cache(self):
        """
        Clear out the content-type cache.
        """
        self._cache.clear()

    def _add_to_cache(self, using, ct):
        """Insert a ContentType into the cache."""
        # Note it's possible for ContentType objects to be stale; model_class() will return None.
        # Hence, there is no reliance on model._meta.app_label here, just using the model fields instead.
        key = (ct.app_label, ct.model)
        self._cache.setdefault(using, {})[key] = ct
        self._cache.setdefault(using, {})[ct.id] = ct


class ContentType(models.Model):
    app_label = models.CharField(max_length=100)
    model = models.CharField(_('python model class name'), max_length=100)
    objects = ContentTypeManager()

    class Meta:
        verbose_name = _('content type')
        verbose_name_plural = _('content types')
        db_table = 'django_content_type'
        unique_together = (('app_label', 'model'),)

    def __str__(self):
        return self.name

    @property
    def name(self):
        model = self.model_class()
        if not model:
            return self.model
        return str(model._meta.verbose_name)

    def model_class(self):
        """Return the model class for this type of content."""
        try:
            return apps.get_model(self.app_label, self.model)
        except LookupError:
            return None

    def get_object_for_this_type(self, **kwargs):
        """
        Return an object of this type for the keyword arguments given.
        Basically, this is a proxy around this object_type's get_object() model
        method. The ObjectNotExist exception, if thrown, will not be caught,
        so code that calls this method should catch it.
        """
        return self.model_class()._base_manager.using(self._state.db).get(**kwargs)

    def get_all_objects_for_this_type(self, **kwargs):
        """
        Return all objects of this type for the keyword arguments given.
        """
        return self.model_class()._base_manager.using(self._state.db).filter(**kwargs)

    def natural_key(self):
        return (self.app_label, self.model)



import jinja2.ext

from ..templatetags.webpack_loader import render_bundle


class WebpackExtension(jinja2.ext.Extension):
    def __init__(self, environment):
        super(WebpackExtension, self).__init__(environment)
        environment.globals["render_bundle"] = lambda *a, **k: jinja2.Markup(render_bundle(*a, **k))



import re

from django.conf import settings


__all__ = ('load_config',)


DEFAULT_CONFIG = {
    'DEFAULT': {
        'CACHE': not settings.DEBUG,
        'BUNDLE_DIR_NAME': 'webpack_bundles/',
        'STATS_FILE': 'webpack-stats.json',
        # FIXME: Explore usage of fsnotify
        'POLL_INTERVAL': 0.1,
        'TIMEOUT': None,
        'IGNORE': ['.+\.hot-update.js', '.+\.map'],
        'LOADER_CLASS': 'webpack_loader.loader.WebpackLoader',
    }
}

user_config = getattr(settings, 'WEBPACK_LOADER', DEFAULT_CONFIG)

user_config = dict(
    (name, dict(DEFAULT_CONFIG['DEFAULT'], **cfg))
    for name, cfg in user_config.items()
)

for entry in user_config.values():
    entry['ignores'] = [re.compile(I) for I in entry['IGNORE']]


def load_config(name):
    return user_config[name]



from django.db.backends.utils import names_digest, split_identifier
from django.db.models.query_utils import Q
from django.db.models.sql import Query

__all__ = ['Index']


class Index:
    suffix = 'idx'
    # The max length of the name of the index (restricted to 30 for
    # cross-database compatibility with Oracle)
    max_name_length = 30

    def __init__(self, *, fields=(), name=None, db_tablespace=None, opclasses=(), condition=None):
        if opclasses and not name:
            raise ValueError('An index must be named to use opclasses.')
        if not isinstance(condition, (type(None), Q)):
            raise ValueError('Index.condition must be a Q instance.')
        if condition and not name:
            raise ValueError('An index must be named to use condition.')
        if not isinstance(fields, (list, tuple)):
            raise ValueError('Index.fields must be a list or tuple.')
        if not isinstance(opclasses, (list, tuple)):
            raise ValueError('Index.opclasses must be a list or tuple.')
        if opclasses and len(fields) != len(opclasses):
            raise ValueError('Index.fields and Index.opclasses must have the same number of elements.')
        if not fields:
            raise ValueError('At least one field is required to define an index.')
        self.fields = list(fields)
        # A list of 2-tuple with the field name and ordering ('' or 'DESC').
        self.fields_orders = [
            (field_name[1:], 'DESC') if field_name.startswith('-') else (field_name, '')
            for field_name in self.fields
        ]
        self.name = name or ''
        if self.name:
            errors = self.check_name()
            if len(self.name) > self.max_name_length:
                errors.append('Index names cannot be longer than %s characters.' % self.max_name_length)
            if errors:
                raise ValueError(errors)
        self.db_tablespace = db_tablespace
        self.opclasses = opclasses
        self.condition = condition

    def check_name(self):
        errors = []
        # Name can't start with an underscore on Oracle; prepend D if needed.
        if self.name[0] == '_':
            errors.append('Index names cannot start with an underscore (_).')
            self.name = 'D%s' % self.name[1:]
        # Name can't start with a number on Oracle; prepend D if needed.
        elif self.name[0].isdigit():
            errors.append('Index names cannot start with a number (0-9).')
            self.name = 'D%s' % self.name[1:]
        return errors

    def _get_condition_sql(self, model, schema_editor):
        if self.condition is None:
            return None
        query = Query(model=model)
        where = query.build_where(self.condition)
        compiler = query.get_compiler(connection=schema_editor.connection)
        sql, params = where.as_sql(compiler, schema_editor.connection)
        return sql % tuple(schema_editor.quote_value(p) for p in params)

    def create_sql(self, model, schema_editor, using=''):
        fields = [model._meta.get_field(field_name) for field_name, _ in self.fields_orders]
        col_suffixes = [order[1] for order in self.fields_orders]
        condition = self._get_condition_sql(model, schema_editor)
        return schema_editor._create_index_sql(
            model, fields, name=self.name, using=using, db_tablespace=self.db_tablespace,
            col_suffixes=col_suffixes, opclasses=self.opclasses, condition=condition,
        )

    def remove_sql(self, model, schema_editor):
        return schema_editor._delete_index_sql(model, self.name)

    def deconstruct(self):
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__name__)
        path = path.replace('django.db.models.indexes', 'django.db.models')
        kwargs = {'fields': self.fields, 'name': self.name}
        if self.db_tablespace is not None:
            kwargs['db_tablespace'] = self.db_tablespace
        if self.opclasses:
            kwargs['opclasses'] = self.opclasses
        if self.condition:
            kwargs['condition'] = self.condition
        return (path, (), kwargs)

    def clone(self):
        """Create a copy of this Index."""
        _, _, kwargs = self.deconstruct()
        return self.__class__(**kwargs)

    def set_name_with_model(self, model):
        """
        Generate a unique name for the index.

        The name is divided into 3 parts - table name (12 chars), field name
        (8 chars) and unique hash + suffix (10 chars). Each part is made to
        fit its size by truncating the excess length.
        """
        _, table_name = split_identifier(model._meta.db_table)
        column_names = [model._meta.get_field(field_name).column for field_name, order in self.fields_orders]
        column_names_with_order = [
            (('-%s' if order else '%s') % column_name)
            for column_name, (field_name, order) in zip(column_names, self.fields_orders)
        ]
        # The length of the parts of the name is based on the default max
        # length of 30 characters.
        hash_data = [table_name] + column_names_with_order + [self.suffix]
        self.name = '%s_%s_%s' % (
            table_name[:11],
            column_names[0][:7],
            '%s_%s' % (names_digest(*hash_data, length=6), self.suffix),
        )
        assert len(self.name) <= self.max_name_length, (
            'Index too long for multiple database support. Is self.suffix '
            'longer than 3 characters?'
        )
        self.check_name()

    def __repr__(self):
        return "<%s: fields='%s'%s>" % (
            self.__class__.__name__, ', '.join(self.fields),
            '' if self.condition is None else ', condition=%s' % self.condition,
        )

    def __eq__(self, other):
        return (self.__class__ == other.__class__) and (self.deconstruct() == other.deconstruct())



import sys

from django.db.backends.base.features import BaseDatabaseFeatures

from .base import Database


class DatabaseFeatures(BaseDatabaseFeatures):
    # SQLite can read from a cursor since SQLite 3.6.5, subject to the caveat
    # that statements within a connection aren't isolated from each other. See
    # https://sqlite.org/isolation.html.
    can_use_chunked_reads = True
    test_db_allows_multiple_connections = False
    supports_unspecified_pk = True
    supports_timezones = False
    max_query_params = 999
    supports_mixed_date_datetime_comparisons = False
    autocommits_when_autocommit_is_off = sys.version_info < (3, 6)
    can_introspect_autofield = True
    can_introspect_decimal_field = False
    can_introspect_duration_field = False
    can_introspect_positive_integer_field = True
    can_introspect_small_integer_field = True
    introspected_big_auto_field_type = 'AutoField'
    supports_transactions = True
    atomic_transactions = False
    can_rollback_ddl = True
    supports_atomic_references_rename = Database.sqlite_version_info >= (3, 26, 0)
    supports_paramstyle_pyformat = False
    supports_sequence_reset = False
    can_clone_databases = True
    supports_temporal_subtraction = True
    ignores_table_name_case = True
    supports_cast_with_precision = False
    time_cast_precision = 3
    can_release_savepoints = True
    # Is "ALTER TABLE ... RENAME COLUMN" supported?
    can_alter_table_rename_column = Database.sqlite_version_info >= (3, 25, 0)
    supports_parentheses_in_compound = False
    # Deferred constraint checks can be emulated on SQLite < 3.20 but not in a
    # reasonably performant way.
    supports_pragma_foreign_key_check = Database.sqlite_version_info >= (3, 20, 0)
    can_defer_constraint_checks = supports_pragma_foreign_key_check
    supports_functions_in_partial_indexes = Database.sqlite_version_info >= (3, 15, 0)



"""Deprecation messages and bits of code used elsewhere in the codebase that
is planned to be removed in the next pytest release.

Keeping it in a central location makes it easy to track what is deprecated and should
be removed when the time comes.

All constants defined in this module should be either instances of
:class:`PytestWarning`, or :class:`UnformattedWarning`
in case of warnings which need to format their messages.
"""
from _pytest.warning_types import PytestDeprecationWarning
from _pytest.warning_types import UnformattedWarning

# set of plugins which have been integrated into the core; we use this list to ignore
# them during registration to avoid conflicts
DEPRECATED_EXTERNAL_PLUGINS = {
    "pytest_catchlog",
    "pytest_capturelog",
    "pytest_faulthandler",
}


FILLFUNCARGS = UnformattedWarning(
    PytestDeprecationWarning,
    "{name} is deprecated, use "
    "function._request._fillfixtures() instead if you cannot avoid reaching into internals.",
)

PYTEST_COLLECT_MODULE = UnformattedWarning(
    PytestDeprecationWarning,
    "pytest.collect.{name} was moved to pytest.{name}\n"
    "Please update to the new name.",
)


MINUS_K_DASH = PytestDeprecationWarning(
    "The `-k '-expr'` syntax to -k is deprecated.\nUse `-k 'not expr'` instead."
)

MINUS_K_COLON = PytestDeprecationWarning(
    "The `-k 'expr:'` syntax to -k is deprecated.\n"
    "Please open an issue if you use this and want a replacement."
)

WARNING_CAPTURED_HOOK = PytestDeprecationWarning(
    "The pytest_warning_captured is deprecated and will be removed in a future release.\n"
    "Please use pytest_warning_recorded instead."
)

FSCOLLECTOR_GETHOOKPROXY_ISINITPATH = PytestDeprecationWarning(
    "The gethookproxy() and isinitpath() methods of FSCollector and Package are deprecated; "
    "use self.session.gethookproxy() and self.session.isinitpath() instead. "
)



#
# The Python Imaging Library.
# $Id$
#
# standard mode descriptors
#
# History:
# 2006-03-20 fl   Added
#
# Copyright (c) 2006 by Secret Labs AB.
# Copyright (c) 2006 by Fredrik Lundh.
#
# See the README file for information on usage and redistribution.
#

# mode descriptor cache
_modes = None


class ModeDescriptor:
    """Wrapper for mode strings."""

    def __init__(self, mode, bands, basemode, basetype):
        self.mode = mode
        self.bands = bands
        self.basemode = basemode
        self.basetype = basetype

    def __str__(self):
        return self.mode


def getmode(mode):
    """Gets a mode descriptor for the given mode."""
    global _modes
    if not _modes:
        # initialize mode cache

        from . import Image

        modes = {}
        # core modes
        for m, (basemode, basetype, bands) in Image._MODEINFO.items():
            modes[m] = ModeDescriptor(m, bands, basemode, basetype)
        # extra experimental modes
        modes["RGBa"] = ModeDescriptor("RGBa", ("R", "G", "B", "a"), "RGB", "L")
        modes["LA"] = ModeDescriptor("LA", ("L", "A"), "L", "L")
        modes["La"] = ModeDescriptor("La", ("L", "a"), "L", "L")
        modes["PA"] = ModeDescriptor("PA", ("P", "A"), "RGB", "L")
        # mapping modes
        for i16mode in (
            "I;16",
            "I;16S",
            "I;16L",
            "I;16LS",
            "I;16B",
            "I;16BS",
            "I;16N",
            "I;16NS",
        ):
            modes[i16mode] = ModeDescriptor(i16mode, ("I",), "L", "L")
        # set global mode cache atomically
        _modes = modes
    return _modes[mode]



from django.utils.cache import (
    cc_delim_re, get_conditional_response, set_response_etag,
)
from django.utils.deprecation import MiddlewareMixin
from django.utils.http import parse_http_date_safe


class ConditionalGetMiddleware(MiddlewareMixin):
    """
    Handle conditional GET operations. If the response has an ETag or
    Last-Modified header and the request has If-None-Match or If-Modified-Since,
    replace the response with HttpNotModified. Add an ETag header if needed.
    """
    def process_response(self, request, response):
        # It's too late to prevent an unsafe request with a 412 response, and
        # for a HEAD request, the response body is always empty so computing
        # an accurate ETag isn't possible.
        if request.method != 'GET':
            return response

        if self.needs_etag(response) and not response.has_header('ETag'):
            set_response_etag(response)

        etag = response.get('ETag')
        last_modified = response.get('Last-Modified')
        last_modified = last_modified and parse_http_date_safe(last_modified)

        if etag or last_modified:
            return get_conditional_response(
                request,
                etag=etag,
                last_modified=last_modified,
                response=response,
            )

        return response

    def needs_etag(self, response):
        """Return True if an ETag header should be added to response."""
        cache_control_headers = cc_delim_re.split(response.get('Cache-Control', ''))
        return all(header.lower() != 'no-store' for header in cache_control_headers)



import os
import warnings
from functools import lru_cache
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union

import py

import _pytest._code
from _pytest._code import getfslineno
from _pytest._code.code import ExceptionInfo
from _pytest._code.code import TerminalRepr
from _pytest.compat import cached_property
from _pytest.compat import overload
from _pytest.compat import TYPE_CHECKING
from _pytest.config import Config
from _pytest.config import ConftestImportFailure
from _pytest.deprecated import FSCOLLECTOR_GETHOOKPROXY_ISINITPATH
from _pytest.fixtures import FixtureDef
from _pytest.fixtures import FixtureLookupError
from _pytest.mark.structures import Mark
from _pytest.mark.structures import MarkDecorator
from _pytest.mark.structures import NodeKeywords
from _pytest.outcomes import fail
from _pytest.pathlib import absolutepath
from _pytest.pathlib import Path
from _pytest.store import Store

if TYPE_CHECKING:
    from typing import Type

    # Imported here due to circular import.
    from _pytest.main import Session
    from _pytest.warning_types import PytestWarning
    from _pytest._code.code import _TracebackStyle


SEP = "/"

tracebackcutdir = py.path.local(_pytest.__file__).dirpath()


@lru_cache(maxsize=None)
def _splitnode(nodeid: str) -> Tuple[str, ...]:
    """Split a nodeid into constituent 'parts'.

    Node IDs are strings, and can be things like:
        ''
        'testing/code'
        'testing/code/test_excinfo.py'
        'testing/code/test_excinfo.py::TestFormattedExcinfo'

    Return values are lists e.g.
        []
        ['testing', 'code']
        ['testing', 'code', 'test_excinfo.py']
        ['testing', 'code', 'test_excinfo.py', 'TestFormattedExcinfo']
    """
    if nodeid == "":
        # If there is no root node at all, return an empty list so the caller's
        # logic can remain sane.
        return ()
    parts = nodeid.split(SEP)
    # Replace single last element 'test_foo.py::Bar' with multiple elements
    # 'test_foo.py', 'Bar'.
    parts[-1:] = parts[-1].split("::")
    # Convert parts into a tuple to avoid possible errors with caching of a
    # mutable type.
    return tuple(parts)


def ischildnode(baseid: str, nodeid: str) -> bool:
    """Return True if the nodeid is a child node of the baseid.

    E.g. 'foo/bar::Baz' is a child of 'foo', 'foo/bar' and 'foo/bar::Baz',
    but not of 'foo/blorp'.
    """
    base_parts = _splitnode(baseid)
    node_parts = _splitnode(nodeid)
    if len(node_parts) < len(base_parts):
        return False
    return node_parts[: len(base_parts)] == base_parts


_NodeType = TypeVar("_NodeType", bound="Node")


class NodeMeta(type):
    def __call__(self, *k, **kw):
        msg = (
            "Direct construction of {name} has been deprecated, please use {name}.from_parent.\n"
            "See "
            "https://docs.pytest.org/en/stable/deprecations.html#node-construction-changed-to-node-from-parent"
            " for more details."
        ).format(name=self.__name__)
        fail(msg, pytrace=False)

    def _create(self, *k, **kw):
        return super().__call__(*k, **kw)


class Node(metaclass=NodeMeta):
    """Base class for Collector and Item, the components of the test
    collection tree.

    Collector subclasses have children; Items are leaf nodes.
    """

    # Use __slots__ to make attribute access faster.
    # Note that __dict__ is still available.
    __slots__ = (
        "name",
        "parent",
        "config",
        "session",
        "fspath",
        "_nodeid",
        "_store",
        "__dict__",
    )

    def __init__(
        self,
        name: str,
        parent: "Optional[Node]" = None,
        config: Optional[Config] = None,
        session: "Optional[Session]" = None,
        fspath: Optional[py.path.local] = None,
        nodeid: Optional[str] = None,
    ) -> None:
        #: A unique name within the scope of the parent node.
        self.name = name

        #: The parent collector node.
        self.parent = parent

        #: The pytest config object.
        if config:
            self.config = config  # type: Config
        else:
            if not parent:
                raise TypeError("config or parent must be provided")
            self.config = parent.config

        #: The pytest session this node is part of.
        if session:
            self.session = session
        else:
            if not parent:
                raise TypeError("session or parent must be provided")
            self.session = parent.session

        #: Filesystem path where this node was collected from (can be None).
        self.fspath = fspath or getattr(parent, "fspath", None)

        #: Keywords/markers collected from all scopes.
        self.keywords = NodeKeywords(self)

        #: The marker objects belonging to this node.
        self.own_markers = []  # type: List[Mark]

        #: Allow adding of extra keywords to use for matching.
        self.extra_keyword_matches = set()  # type: Set[str]

        # Used for storing artificial fixturedefs for direct parametrization.
        self._name2pseudofixturedef = {}  # type: Dict[str, FixtureDef[Any]]

        if nodeid is not None:
            assert "::()" not in nodeid
            self._nodeid = nodeid
        else:
            if not self.parent:
                raise TypeError("nodeid or parent must be provided")
            self._nodeid = self.parent.nodeid
            if self.name != "()":
                self._nodeid += "::" + self.name

        # A place where plugins can store information on the node for their
        # own use. Currently only intended for internal plugins.
        self._store = Store()

    @classmethod
    def from_parent(cls, parent: "Node", **kw):
        """Public constructor for Nodes.

        This indirection got introduced in order to enable removing
        the fragile logic from the node constructors.

        Subclasses can use ``super().from_parent(...)`` when overriding the
        construction.

        :param parent: The parent node of this Node.
        """
        if "config" in kw:
            raise TypeError("config is not a valid argument for from_parent")
        if "session" in kw:
            raise TypeError("session is not a valid argument for from_parent")
        return cls._create(parent=parent, **kw)

    @property
    def ihook(self):
        """fspath-sensitive hook proxy used to call pytest hooks."""
        return self.session.gethookproxy(self.fspath)

    def __repr__(self) -> str:
        return "<{} {}>".format(self.__class__.__name__, getattr(self, "name", None))

    def warn(self, warning: "PytestWarning") -> None:
        """Issue a warning for this Node.

        Warnings will be displayed after the test session, unless explicitly suppressed.

        :param Warning warning:
            The warning instance to issue. Must be a subclass of PytestWarning.

        :raises ValueError: If ``warning`` instance is not a subclass of PytestWarning.

        Example usage:

        .. code-block:: python

            node.warn(PytestWarning("some message"))
        """
        from _pytest.warning_types import PytestWarning

        if not isinstance(warning, PytestWarning):
            raise ValueError(
                "warning must be an instance of PytestWarning or subclass, got {!r}".format(
                    warning
                )
            )
        path, lineno = get_fslocation_from_item(self)
        assert lineno is not None
        warnings.warn_explicit(
            warning, category=None, filename=str(path), lineno=lineno + 1,
        )

    # Methods for ordering nodes.

    @property
    def nodeid(self) -> str:
        """A ::-separated string denoting its collection tree address."""
        return self._nodeid

    def __hash__(self) -> int:
        return hash(self._nodeid)

    def setup(self) -> None:
        pass

    def teardown(self) -> None:
        pass

    def listchain(self) -> List["Node"]:
        """Return list of all parent collectors up to self, starting from
        the root of collection tree."""
        chain = []
        item = self  # type: Optional[Node]
        while item is not None:
            chain.append(item)
            item = item.parent
        chain.reverse()
        return chain

    def add_marker(
        self, marker: Union[str, MarkDecorator], append: bool = True
    ) -> None:
        """Dynamically add a marker object to the node.

        :param append:
            Whether to append the marker, or prepend it.
        """
        from _pytest.mark import MARK_GEN

        if isinstance(marker, MarkDecorator):
            marker_ = marker
        elif isinstance(marker, str):
            marker_ = getattr(MARK_GEN, marker)
        else:
            raise ValueError("is not a string or pytest.mark.* Marker")
        self.keywords[marker_.name] = marker_
        if append:
            self.own_markers.append(marker_.mark)
        else:
            self.own_markers.insert(0, marker_.mark)

    def iter_markers(self, name: Optional[str] = None) -> Iterator[Mark]:
        """Iterate over all markers of the node.

        :param name: If given, filter the results by the name attribute.
        """
        return (x[1] for x in self.iter_markers_with_node(name=name))

    def iter_markers_with_node(
        self, name: Optional[str] = None
    ) -> Iterator[Tuple["Node", Mark]]:
        """Iterate over all markers of the node.

        :param name: If given, filter the results by the name attribute.
        :returns: An iterator of (node, mark) tuples.
        """
        for node in reversed(self.listchain()):
            for mark in node.own_markers:
                if name is None or getattr(mark, "name", None) == name:
                    yield node, mark

    @overload
    def get_closest_marker(self, name: str) -> Optional[Mark]:
        ...

    @overload  # noqa: F811
    def get_closest_marker(self, name: str, default: Mark) -> Mark:  # noqa: F811
        ...

    def get_closest_marker(  # noqa: F811
        self, name: str, default: Optional[Mark] = None
    ) -> Optional[Mark]:
        """Return the first marker matching the name, from closest (for
        example function) to farther level (for example module level).

        :param default: Fallback return value if no marker was found.
        :param name: Name to filter by.
        """
        return next(self.iter_markers(name=name), default)

    def listextrakeywords(self) -> Set[str]:
        """Return a set of all extra keywords in self and any parents."""
        extra_keywords = set()  # type: Set[str]
        for item in self.listchain():
            extra_keywords.update(item.extra_keyword_matches)
        return extra_keywords

    def listnames(self) -> List[str]:
        return [x.name for x in self.listchain()]

    def addfinalizer(self, fin: Callable[[], object]) -> None:
        """Register a function to be called when this node is finalized.

        This method can only be called when this node is active
        in a setup chain, for example during self.setup().
        """
        self.session._setupstate.addfinalizer(fin, self)

    def getparent(self, cls: "Type[_NodeType]") -> Optional[_NodeType]:
        """Get the next parent node (including self) which is an instance of
        the given class."""
        current = self  # type: Optional[Node]
        while current and not isinstance(current, cls):
            current = current.parent
        assert current is None or isinstance(current, cls)
        return current

    def _prunetraceback(self, excinfo: ExceptionInfo[BaseException]) -> None:
        pass

    def _repr_failure_py(
        self,
        excinfo: ExceptionInfo[BaseException],
        style: "Optional[_TracebackStyle]" = None,
    ) -> TerminalRepr:
        if isinstance(excinfo.value, ConftestImportFailure):
            excinfo = ExceptionInfo(excinfo.value.excinfo)
        if isinstance(excinfo.value, fail.Exception):
            if not excinfo.value.pytrace:
                style = "value"
        if isinstance(excinfo.value, FixtureLookupError):
            return excinfo.value.formatrepr()
        if self.config.getoption("fulltrace", False):
            style = "long"
        else:
            tb = _pytest._code.Traceback([excinfo.traceback[-1]])
            self._prunetraceback(excinfo)
            if len(excinfo.traceback) == 0:
                excinfo.traceback = tb
            if style == "auto":
                style = "long"
        # XXX should excinfo.getrepr record all data and toterminal() process it?
        if style is None:
            if self.config.getoption("tbstyle", "auto") == "short":
                style = "short"
            else:
                style = "long"

        if self.config.getoption("verbose", 0) > 1:
            truncate_locals = False
        else:
            truncate_locals = True

        # excinfo.getrepr() formats paths relative to the CWD if `abspath` is False.
        # It is possible for a fixture/test to change the CWD while this code runs, which
        # would then result in the user seeing confusing paths in the failure message.
        # To fix this, if the CWD changed, always display the full absolute path.
        # It will be better to just always display paths relative to invocation_dir, but
        # this requires a lot of plumbing (#6428).
        try:
            abspath = Path(os.getcwd()) != self.config.invocation_params.dir
        except OSError:
            abspath = True

        return excinfo.getrepr(
            funcargs=True,
            abspath=abspath,
            showlocals=self.config.getoption("showlocals", False),
            style=style,
            tbfilter=False,  # pruned already, or in --fulltrace mode.
            truncate_locals=truncate_locals,
        )

    def repr_failure(
        self,
        excinfo: ExceptionInfo[BaseException],
        style: "Optional[_TracebackStyle]" = None,
    ) -> Union[str, TerminalRepr]:
        """Return a representation of a collection or test failure.

        :param excinfo: Exception information for the failure.
        """
        return self._repr_failure_py(excinfo, style)


def get_fslocation_from_item(
    node: "Node",
) -> Tuple[Union[str, py.path.local], Optional[int]]:
    """Try to extract the actual location from a node, depending on available attributes:

    * "location": a pair (path, lineno)
    * "obj": a Python object that the node wraps.
    * "fspath": just a path

    :rtype: A tuple of (str|py.path.local, int) with filename and line number.
    """
    # See Item.location.
    location = getattr(
        node, "location", None
    )  # type: Optional[Tuple[str, Optional[int], str]]
    if location is not None:
        return location[:2]
    obj = getattr(node, "obj", None)
    if obj is not None:
        return getfslineno(obj)
    return getattr(node, "fspath", "unknown location"), -1


class Collector(Node):
    """Collector instances create children through collect() and thus
    iteratively build a tree."""

    class CollectError(Exception):
        """An error during collection, contains a custom message."""

    def collect(self) -> Iterable[Union["Item", "Collector"]]:
        """Return a list of children (items and collectors) for this
        collection node."""
        raise NotImplementedError("abstract")

    # TODO: This omits the style= parameter which breaks Liskov Substitution.
    def repr_failure(  # type: ignore[override]
        self, excinfo: ExceptionInfo[BaseException]
    ) -> Union[str, TerminalRepr]:
        """Return a representation of a collection failure.

        :param excinfo: Exception information for the failure.
        """
        if isinstance(excinfo.value, self.CollectError) and not self.config.getoption(
            "fulltrace", False
        ):
            exc = excinfo.value
            return str(exc.args[0])

        # Respect explicit tbstyle option, but default to "short"
        # (_repr_failure_py uses "long" with "fulltrace" option always).
        tbstyle = self.config.getoption("tbstyle", "auto")
        if tbstyle == "auto":
            tbstyle = "short"

        return self._repr_failure_py(excinfo, style=tbstyle)

    def _prunetraceback(self, excinfo: ExceptionInfo[BaseException]) -> None:
        if hasattr(self, "fspath"):
            traceback = excinfo.traceback
            ntraceback = traceback.cut(path=self.fspath)
            if ntraceback == traceback:
                ntraceback = ntraceback.cut(excludepath=tracebackcutdir)
            excinfo.traceback = ntraceback.filter()


def _check_initialpaths_for_relpath(session, fspath):
    for initial_path in session._initialpaths:
        if fspath.common(initial_path) == initial_path:
            return fspath.relto(initial_path)


class FSCollector(Collector):
    def __init__(
        self,
        fspath: py.path.local,
        parent=None,
        config: Optional[Config] = None,
        session: Optional["Session"] = None,
        nodeid: Optional[str] = None,
    ) -> None:
        name = fspath.basename
        if parent is not None:
            rel = fspath.relto(parent.fspath)
            if rel:
                name = rel
            name = name.replace(os.sep, SEP)
        self.fspath = fspath

        session = session or parent.session

        if nodeid is None:
            nodeid = self.fspath.relto(session.config.rootdir)

            if not nodeid:
                nodeid = _check_initialpaths_for_relpath(session, fspath)
            if nodeid and os.sep != SEP:
                nodeid = nodeid.replace(os.sep, SEP)

        super().__init__(name, parent, config, session, nodeid=nodeid, fspath=fspath)

    @classmethod
    def from_parent(cls, parent, *, fspath, **kw):
        """The public constructor."""
        return super().from_parent(parent=parent, fspath=fspath, **kw)

    def gethookproxy(self, fspath: py.path.local):
        warnings.warn(FSCOLLECTOR_GETHOOKPROXY_ISINITPATH, stacklevel=2)
        return self.session.gethookproxy(fspath)

    def isinitpath(self, path: py.path.local) -> bool:
        warnings.warn(FSCOLLECTOR_GETHOOKPROXY_ISINITPATH, stacklevel=2)
        return self.session.isinitpath(path)


class File(FSCollector):
    """Base class for collecting tests from a file.

    :ref:`non-python tests`.
    """


class Item(Node):
    """A basic test invocation item.

    Note that for a single function there might be multiple test invocation items.
    """

    nextitem = None

    def __init__(
        self,
        name,
        parent=None,
        config: Optional[Config] = None,
        session: Optional["Session"] = None,
        nodeid: Optional[str] = None,
    ) -> None:
        super().__init__(name, parent, config, session, nodeid=nodeid)
        self._report_sections = []  # type: List[Tuple[str, str, str]]

        #: A list of tuples (name, value) that holds user defined properties
        #: for this test.
        self.user_properties = []  # type: List[Tuple[str, object]]

    def runtest(self) -> None:
        raise NotImplementedError("runtest must be implemented by Item subclass")

    def add_report_section(self, when: str, key: str, content: str) -> None:
        """Add a new report section, similar to what's done internally to add
        stdout and stderr captured output::

            item.add_report_section("call", "stdout", "report section contents")

        :param str when:
            One of the possible capture states, ``"setup"``, ``"call"``, ``"teardown"``.
        :param str key:
            Name of the section, can be customized at will. Pytest uses ``"stdout"`` and
            ``"stderr"`` internally.
        :param str content:
            The full contents as a string.
        """
        if content:
            self._report_sections.append((when, key, content))

    def reportinfo(self) -> Tuple[Union[py.path.local, str], Optional[int], str]:
        return self.fspath, None, ""

    @cached_property
    def location(self) -> Tuple[str, Optional[int], str]:
        location = self.reportinfo()
        fspath = absolutepath(str(location[0]))
        relfspath = self.session._node_location_to_relpath(fspath)
        assert type(location[2]) is str
        return (relfspath, location[1], location[2])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2009-2018 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""Python 2/3 compatibility.

This module only exists to avoid a dependency on six
for very trivial stuff. We only need to take care of
string types, buffers and metaclasses.

Parts of the code is copied directly from six:
https://bitbucket.org/gutworth/six
"""

import sys
from io import TextIOBase

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3


if PY3:
    def unicode_compatible(cls):
        return cls

    text_type = str
    string_types = (str,)
    from io import StringIO
    file_types = (StringIO, TextIOBase)


elif PY2:
    def unicode_compatible(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode('utf-8')
        return cls

    text_type = unicode
    string_types = (str, unicode,)
    from StringIO import StringIO
    file_types = (file, StringIO, TextIOBase)



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

"""SQL Lexer"""

# This code is based on the SqlLexer in pygments.
# http://pygments.org/
# It's separated from the rest of pygments to increase performance
# and to allow some customizations.

from io import TextIOBase

from sqlparse import tokens
from sqlparse.keywords import SQL_REGEX
from sqlparse.utils import consume


class Lexer:
    """Lexer
    Empty class. Leaving for backwards-compatibility
    """

    @staticmethod
    def get_tokens(text, encoding=None):
        """
        Return an iterable of (tokentype, value) pairs generated from
        `text`. If `unfiltered` is set to `True`, the filtering mechanism
        is bypassed even if filters are defined.

        Also preprocess the text, i.e. expand tabs and strip it if
        wanted and applies registered filters.

        Split ``text`` into (tokentype, text) pairs.

        ``stack`` is the initial stack (default: ``['root']``)
        """
        if isinstance(text, TextIOBase):
            text = text.read()

        if isinstance(text, str):
            pass
        elif isinstance(text, bytes):
            if encoding:
                text = text.decode(encoding)
            else:
                try:
                    text = text.decode('utf-8')
                except UnicodeDecodeError:
                    text = text.decode('unicode-escape')
        else:
            raise TypeError("Expected text or file-like object, got {!r}".
                            format(type(text)))

        iterable = enumerate(text)
        for pos, char in iterable:
            for rexmatch, action in SQL_REGEX:
                m = rexmatch(text, pos)

                if not m:
                    continue
                elif isinstance(action, tokens._TokenType):
                    yield action, m.group()
                elif callable(action):
                    yield action(m.group())

                consume(iterable, m.end() - pos - 1)
                break
            else:
                yield tokens.Error, char


def tokenize(sql, encoding=None):
    """Tokenize sql.

    Tokenize *sql* using the :class:`Lexer` and return a 2-tuple stream
    of ``(token type, value)`` items.
    """
    return Lexer().get_tokens(sql, encoding)



# Generated by Django 2.2.11 on 2021-02-24 00:19

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pagebuilder', '0003_auto_20210128_0004'),
    ]

    operations = [
        migrations.RenameField(
            model_name='page_version',
            old_name='description',
            new_name='seo_description',
        ),
    ]



from django.db import models
from django.db.models.query import QuerySet
from django.urls import reverse
from django.utils import timezone
from django.utils.text import slugify

from scarlet.assets.fields import AssetsFileField
from scarlet.assets.models import Asset
from scarlet.cms.fields import HTMLTextField, OrderField
from scarlet.versioning.models import Cloneable

from . import settings


class BasePage(models.Model):
    """
    Page fields contain metadata and SEO elements for the page.
    """

    _clone_related = []  # must be set by implementing class
    versioned_unique = ("slug",)

    title = models.CharField(max_length=255)
    internal_name = models.CharField(
        max_length=255, blank=True, null=True, help_text="For internal use only."
    )
    slug = models.SlugField(max_length=255)
    seo_description = models.CharField(max_length=500, blank=True, null=True)
    keywords = models.CharField(max_length=255, blank=True, null=True)
    og_image = AssetsFileField(
        type=Asset.IMAGE, blank=True, null=True, verbose_name="Open Graph Image"
    )

    created = models.DateTimeField(default=timezone.now, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)

    def __str__(self) -> str:
        return self.internal_name if self.internal_name else self.title

    def get_absolute_url(self):
        return reverse("page-details", kwargs={"slug": self.slug})

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.name)
        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class AbstractSupportingModule(Cloneable):
    truncate_at = 100
    order = OrderField()
    internal_name = models.CharField(max_length=255, blank=True, null=True)
    style = models.CharField(max_length=255, blank=True, null=True, choices=())
    behavior = models.CharField(max_length=255, blank=True, null=True, choices=())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if hasattr(self, 'styles'):
            self._meta.get_field('style').choices = self.styles
        else:
            self._meta.get_field('style').editable = False

        if hasattr(self, 'behaviors'):
            self._meta.get_field('behavior').choices = self.behaviors
        else:
            self._meta.get_field('behavior').editable = False

    @property
    def kind(self) -> str:
        raise NotImplementedError

    def _truncate_for_display(self, str: str) -> str:
        response = str
        if len(str) > self.truncate_at:
            response = f"{str[:self.truncate_at]}…"
        return response

    @property
    def display_name(self) -> str:
        display_name = "No Display Name Set"
        if self.internal_name:
            display_name = self.internal_name
        elif hasattr(self, "title"):
            display_name = self.title
        return display_name

    def __str__(self) -> str:
        return f"{self.display_name} ({self.kind} Module)"

    class Meta:
        abstract = True


class BaseHeroModule(AbstractSupportingModule):
    image = AssetsFileField(
        type=Asset.IMAGE,
        required_tags=("hero",),
        image_sizes=settings.WIDE_IMAGE_SIZES,
        blank=True,
        null=True,
    )
    header = models.CharField(
        max_length=255, blank=True, null=True, verbose_name="Header Text"
    )
    subheader = models.CharField(
        max_length=255, blank=True, null=True, verbose_name="Subheader Text"
    )
    button_url = models.URLField(blank=True, null=True)
    button_text = models.CharField(max_length=100, blank=True, null=True)

    @property
    def kind(self) -> str:
        return settings.HERO_MODULE


class BaseTwoColumnModule(AbstractSupportingModule):
    left_image = AssetsFileField(
        type=Asset.IMAGE,
        required_tags=("hero",),
        image_sizes=settings.SQUARE_IMAGE_SIZES,
        blank=True,
        null=True,
    )
    left_header = models.CharField(
        max_length=255, blank=True, null=True, verbose_name="Header Text"
    )
    left_description = HTMLTextField(blank=True, null=True, config="useMinimal")
    left_button_url = models.URLField(blank=True, null=True)
    left_button_text = models.CharField(max_length=100, blank=True, null=True)
    right_image = AssetsFileField(
        type=Asset.IMAGE,
        required_tags=("hero",),
        image_sizes=settings.SQUARE_IMAGE_SIZES,
        blank=True,
        null=True,
    )
    right_header = models.CharField(
        max_length=255, blank=True, null=True, verbose_name="Header Text"
    )
    right_description = HTMLTextField(blank=True, null=True, config="useMinimal")
    right_button_url = models.URLField(blank=True, null=True)
    right_button_text = models.CharField(max_length=100, blank=True, null=True)

    @property
    def kind(self) -> str:
        return settings.TWO_COLUMN_MODULE


class BaseIconListManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return (
            super(BaseIconList, self)
            .get_query_set()
            .prefetch_related("iconitem_set")
        )


class BaseIconListModule(AbstractSupportingModule):
    title = models.CharField(max_length=255)
    subtitle = models.CharField(max_length=255, blank=True, null=True)
    footnote = models.CharField(max_length=255, blank=True, null=True)

    objects = BaseIconListManager()

    @property
    def items(self) -> QuerySet:
        # Does this need to be more generic?
        return self.iconlistitem_set.all()

    @property
    def kind(self) -> str:
        return settings.ICON_LIST_MODULE


class BaseIconListItem(models.Model):
    icon = AssetsFileField(type=Asset.SVG_IMAGE)
    text = HTMLTextField(blank=True, null=True, config="useMinimal")
    order = OrderField(default=0)

    @property
    def kind(self) -> str:
        return settings.ICON_LIST_ITEM


class BaseHeaderModule(AbstractSupportingModule):
    header = models.CharField(
        max_length=255, blank=True, null=True, verbose_name="Header Text"
    )
    subheader = models.CharField(
        max_length=255, blank=True, null=True, verbose_name="Subheader Text"
    )

    @property
    def kind(self) -> str:
        return settings.HEADER_MODULE


class BaseFAQManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return (
            super(BaseFAQModule, self)
            .get_query_set()
            .prefetch_related("faqitem_set")
        )


class BaseFAQModule(AbstractSupportingModule):
    objects = BaseFAQManager()

    @property
    def items(self) -> QuerySet:
        return self.faqitem_set.all()

    @property
    def kind(self) -> str:
        return settings.FAQ_MODULE


class BaseFAQItem(AbstractSupportingModule):
    question = models.CharField(max_length=255)
    answer = HTMLTextField(config="useMinimal")

    @property
    def kind(self) -> str:
        return settings.FAQ_ITEM


class BaseLocationManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return (
            super(BaseLocationModule, self)
            .get_query_set()
            .prefetch_related("locationitem_set")
        )


class BaseLocationModule(AbstractSupportingModule):
    objects = BaseFAQManager()

    @property
    def items(self) -> QuerySet:
        return self.locationitem_set.all()

    @property
    def kind(self) -> str:
        return settings.LOCATION_MODULE


class BaseLocationItem(AbstractSupportingModule):
    title = models.CharField(max_length=255)
    address1 = models.CharField(max_length=255, help_text="Street address, P.O. Box")
    address2 = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text="Apartment, suite, unit, building, floor, etc",
    )
    city = models.CharField(max_length=255, blank=True, null=True)
    state = models.CharField(
        max_length=255, verbose_name="State/Province/Region", blank=True, null=True
    )
    zip_code = models.CharField(
        max_length=10, verbose_name="Zip/Postal Code", blank=True, null=True
    )
    country = models.CharField(max_length=100, blank=True, null=True)
    hours = HTMLTextField(config="useMinimal", blank=True, null=True)
    phone_number = models.CharField(max_length=20, blank=True, null=True)

    @property
    def kind(self) -> str:
        return settings.LOCATION_ITEM


class BaseGalleryManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return (
            super(BaseImageGalleryModule, self)
            .get_query_set()
            .prefetch_related("galleryimage_set")
        )


class BaseImageGalleryModule(AbstractSupportingModule):
    objects = BaseGalleryManager()

    @property
    def items(self) -> QuerySet:
        return self.galleryimage_set.all()

    @property
    def kind(self) -> str:
        return settings.GALLERY_MODULE


class BaseGalleryImage(AbstractSupportingModule):
    image = AssetsFileField(
        type=Asset.IMAGE,
        required_tags=("gallery image",),
        image_sizes=settings.GALLERY_IMAGE_SIZES,
        blank=True,
        null=True,
    )

    @property
    def kind(self) -> str:
        return settings.GALLERY_IMAGE


class BaseCarouselManager(models.Manager):
    use_for_related_fields = True

    def get_query_set(self) -> QuerySet:
        return (
            super(BaseCarouselModule, self)
            .get_query_set()
            .prefetch_related("carouselitem_set")
        )


class BaseCarouselModule(AbstractSupportingModule):
    objects = BaseCarouselManager()

    button_url = models.URLField(blank=True, null=True)
    button_text = models.CharField(max_length=100, blank=True, null=True)

    @property
    def items(self) -> QuerySet:
        return self.carouselitem_set.all()

    @property
    def kind(self) -> str:
        return settings.CAROUSEL_MODULE


class BaseCarouselItem(AbstractSupportingModule):
    title = models.CharField(max_length=255)
    subtitle = models.CharField(max_length=255, blank=True, null=True)
    image = AssetsFileField(
        type=Asset.IMAGE,
        required_tags=("carousel image",),
        image_sizes=settings.CAROUSEL_IMAGE_SIZES,
        blank=True,
        null=True,
    )
    button_url = models.URLField(blank=True, null=True)
    button_text = models.CharField(max_length=100, blank=True, null=True)

    @property
    def kind(self) -> str:
        return settings.CAROUSEL_ITEM



"""Prepares a distribution for installation
"""

# The following comment should be removed at some point in the future.
# mypy: strict-optional=False

import logging
import mimetypes
import os
import shutil
import sys

from pip._vendor import requests
from pip._vendor.six import PY2

from pip._internal.distributions import (
    make_distribution_for_install_requirement,
)
from pip._internal.distributions.installed import InstalledDistribution
from pip._internal.exceptions import (
    DirectoryUrlHashUnsupported,
    HashMismatch,
    HashUnpinned,
    InstallationError,
    PreviousBuildDirError,
    VcsHashUnsupported,
)
from pip._internal.utils.filesystem import copy2_fixed
from pip._internal.utils.hashes import MissingHashes
from pip._internal.utils.logging import indent_log
from pip._internal.utils.marker_files import write_delete_marker_file
from pip._internal.utils.misc import (
    ask_path_exists,
    backup_dir,
    display_path,
    hide_url,
    path_to_display,
    rmtree,
)
from pip._internal.utils.temp_dir import TempDirectory
from pip._internal.utils.typing import MYPY_CHECK_RUNNING
from pip._internal.utils.unpacking import unpack_file
from pip._internal.vcs import vcs

if MYPY_CHECK_RUNNING:
    from typing import (
        Callable, List, Optional, Tuple,
    )

    from mypy_extensions import TypedDict

    from pip._internal.distributions import AbstractDistribution
    from pip._internal.index.package_finder import PackageFinder
    from pip._internal.models.link import Link
    from pip._internal.network.download import Downloader
    from pip._internal.req.req_install import InstallRequirement
    from pip._internal.req.req_tracker import RequirementTracker
    from pip._internal.utils.hashes import Hashes

    if PY2:
        CopytreeKwargs = TypedDict(
            'CopytreeKwargs',
            {
                'ignore': Callable[[str, List[str]], List[str]],
                'symlinks': bool,
            },
            total=False,
        )
    else:
        CopytreeKwargs = TypedDict(
            'CopytreeKwargs',
            {
                'copy_function': Callable[[str, str], None],
                'ignore': Callable[[str, List[str]], List[str]],
                'ignore_dangling_symlinks': bool,
                'symlinks': bool,
            },
            total=False,
        )

logger = logging.getLogger(__name__)


def _get_prepared_distribution(
        req,  # type: InstallRequirement
        req_tracker,  # type: RequirementTracker
        finder,  # type: PackageFinder
        build_isolation  # type: bool
):
    # type: (...) -> AbstractDistribution
    """Prepare a distribution for installation.
    """
    abstract_dist = make_distribution_for_install_requirement(req)
    with req_tracker.track(req):
        abstract_dist.prepare_distribution_metadata(finder, build_isolation)
    return abstract_dist


def unpack_vcs_link(link, location):
    # type: (Link, str) -> None
    vcs_backend = vcs.get_backend_for_scheme(link.scheme)
    assert vcs_backend is not None
    vcs_backend.unpack(location, url=hide_url(link.url))


def _copy_file(filename, location, link):
    # type: (str, str, Link) -> None
    copy = True
    download_location = os.path.join(location, link.filename)
    if os.path.exists(download_location):
        response = ask_path_exists(
            'The file {} exists. (i)gnore, (w)ipe, (b)ackup, (a)abort'.format(
                display_path(download_location)
            ),
            ('i', 'w', 'b', 'a'),
        )
        if response == 'i':
            copy = False
        elif response == 'w':
            logger.warning('Deleting %s', display_path(download_location))
            os.remove(download_location)
        elif response == 'b':
            dest_file = backup_dir(download_location)
            logger.warning(
                'Backing up %s to %s',
                display_path(download_location),
                display_path(dest_file),
            )
            shutil.move(download_location, dest_file)
        elif response == 'a':
            sys.exit(-1)
    if copy:
        shutil.copy(filename, download_location)
        logger.info('Saved %s', display_path(download_location))


def unpack_http_url(
    link,  # type: Link
    location,  # type: str
    downloader,  # type: Downloader
    download_dir=None,  # type: Optional[str]
    hashes=None,  # type: Optional[Hashes]
):
    # type: (...) -> str
    temp_dir = TempDirectory(kind="unpack", globally_managed=True)
    # If a download dir is specified, is the file already downloaded there?
    already_downloaded_path = None
    if download_dir:
        already_downloaded_path = _check_download_dir(
            link, download_dir, hashes
        )

    if already_downloaded_path:
        from_path = already_downloaded_path
        content_type = mimetypes.guess_type(from_path)[0]
    else:
        # let's download to a tmp dir
        from_path, content_type = _download_http_url(
            link, downloader, temp_dir.path, hashes
        )

    # unpack the archive to the build dir location. even when only
    # downloading archives, they have to be unpacked to parse dependencies
    unpack_file(from_path, location, content_type)

    return from_path


def _copy2_ignoring_special_files(src, dest):
    # type: (str, str) -> None
    """Copying special files is not supported, but as a convenience to users
    we skip errors copying them. This supports tools that may create e.g.
    socket files in the project source directory.
    """
    try:
        copy2_fixed(src, dest)
    except shutil.SpecialFileError as e:
        # SpecialFileError may be raised due to either the source or
        # destination. If the destination was the cause then we would actually
        # care, but since the destination directory is deleted prior to
        # copy we ignore all of them assuming it is caused by the source.
        logger.warning(
            "Ignoring special file error '%s' encountered copying %s to %s.",
            str(e),
            path_to_display(src),
            path_to_display(dest),
        )


def _copy_source_tree(source, target):
    # type: (str, str) -> None
    def ignore(d, names):
        # type: (str, List[str]) -> List[str]
        # Pulling in those directories can potentially be very slow,
        # exclude the following directories if they appear in the top
        # level dir (and only it).
        # See discussion at https://github.com/pypa/pip/pull/6770
        return ['.tox', '.nox'] if d == source else []

    kwargs = dict(ignore=ignore, symlinks=True)  # type: CopytreeKwargs

    if not PY2:
        # Python 2 does not support copy_function, so we only ignore
        # errors on special file copy in Python 3.
        kwargs['copy_function'] = _copy2_ignoring_special_files

    shutil.copytree(source, target, **kwargs)


def unpack_file_url(
    link,  # type: Link
    location,  # type: str
    download_dir=None,  # type: Optional[str]
    hashes=None  # type: Optional[Hashes]
):
    # type: (...) -> Optional[str]
    """Unpack link into location.
    """
    link_path = link.file_path
    # If it's a url to a local directory
    if link.is_existing_dir():
        if os.path.isdir(location):
            rmtree(location)
        _copy_source_tree(link_path, location)
        return None

    # If a download dir is specified, is the file already there and valid?
    already_downloaded_path = None
    if download_dir:
        already_downloaded_path = _check_download_dir(
            link, download_dir, hashes
        )

    if already_downloaded_path:
        from_path = already_downloaded_path
    else:
        from_path = link_path

    # If --require-hashes is off, `hashes` is either empty, the
    # link's embedded hash, or MissingHashes; it is required to
    # match. If --require-hashes is on, we are satisfied by any
    # hash in `hashes` matching: a URL-based or an option-based
    # one; no internet-sourced hash will be in `hashes`.
    if hashes:
        hashes.check_against_path(from_path)

    content_type = mimetypes.guess_type(from_path)[0]

    # unpack the archive to the build dir location. even when only downloading
    # archives, they have to be unpacked to parse dependencies
    unpack_file(from_path, location, content_type)

    return from_path


def unpack_url(
    link,  # type: Link
    location,  # type: str
    downloader,  # type: Downloader
    download_dir=None,  # type: Optional[str]
    hashes=None,  # type: Optional[Hashes]
):
    # type: (...) -> Optional[str]
    """Unpack link into location, downloading if required.

    :param hashes: A Hashes object, one of whose embedded hashes must match,
        or HashMismatch will be raised. If the Hashes is empty, no matches are
        required, and unhashable types of requirements (like VCS ones, which
        would ordinarily raise HashUnsupported) are allowed.
    """
    # non-editable vcs urls
    if link.is_vcs:
        unpack_vcs_link(link, location)
        return None

    # file urls
    elif link.is_file:
        return unpack_file_url(link, location, download_dir, hashes=hashes)

    # http urls
    else:
        return unpack_http_url(
            link,
            location,
            downloader,
            download_dir,
            hashes=hashes,
        )


def _download_http_url(
    link,  # type: Link
    downloader,  # type: Downloader
    temp_dir,  # type: str
    hashes,  # type: Optional[Hashes]
):
    # type: (...) -> Tuple[str, str]
    """Download link url into temp_dir using provided session"""
    download = downloader(link)

    file_path = os.path.join(temp_dir, download.filename)
    with open(file_path, 'wb') as content_file:
        for chunk in download.chunks:
            content_file.write(chunk)

    if hashes:
        hashes.check_against_path(file_path)

    return file_path, download.response.headers.get('content-type', '')


def _check_download_dir(link, download_dir, hashes):
    # type: (Link, str, Optional[Hashes]) -> Optional[str]
    """ Check download_dir for previously downloaded file with correct hash
        If a correct file is found return its path else None
    """
    download_path = os.path.join(download_dir, link.filename)

    if not os.path.exists(download_path):
        return None

    # If already downloaded, does its hash match?
    logger.info('File was already downloaded %s', download_path)
    if hashes:
        try:
            hashes.check_against_path(download_path)
        except HashMismatch:
            logger.warning(
                'Previously-downloaded file %s has bad hash. '
                'Re-downloading.',
                download_path
            )
            os.unlink(download_path)
            return None
    return download_path


class RequirementPreparer(object):
    """Prepares a Requirement
    """

    def __init__(
        self,
        build_dir,  # type: str
        download_dir,  # type: Optional[str]
        src_dir,  # type: str
        wheel_download_dir,  # type: Optional[str]
        build_isolation,  # type: bool
        req_tracker,  # type: RequirementTracker
        downloader,  # type: Downloader
        finder,  # type: PackageFinder
        require_hashes,  # type: bool
        use_user_site,  # type: bool
    ):
        # type: (...) -> None
        super(RequirementPreparer, self).__init__()

        self.src_dir = src_dir
        self.build_dir = build_dir
        self.req_tracker = req_tracker
        self.downloader = downloader
        self.finder = finder

        # Where still-packed archives should be written to. If None, they are
        # not saved, and are deleted immediately after unpacking.
        self.download_dir = download_dir

        # Where still-packed .whl files should be written to. If None, they are
        # written to the download_dir parameter. Separate to download_dir to
        # permit only keeping wheel archives for pip wheel.
        self.wheel_download_dir = wheel_download_dir

        # NOTE
        # download_dir and wheel_download_dir overlap semantically and may
        # be combined if we're willing to have non-wheel archives present in
        # the wheelhouse output by 'pip wheel'.

        # Is build isolation allowed?
        self.build_isolation = build_isolation

        # Should hash-checking be required?
        self.require_hashes = require_hashes

        # Should install in user site-packages?
        self.use_user_site = use_user_site

    @property
    def _download_should_save(self):
        # type: () -> bool
        if not self.download_dir:
            return False

        if os.path.exists(self.download_dir):
            return True

        logger.critical('Could not find download directory')
        raise InstallationError(
            "Could not find or access download directory '{}'"
            .format(self.download_dir))

    def prepare_linked_requirement(
        self,
        req,  # type: InstallRequirement
    ):
        # type: (...) -> AbstractDistribution
        """Prepare a requirement that would be obtained from req.link
        """
        assert req.link
        link = req.link

        # TODO: Breakup into smaller functions
        if link.scheme == 'file':
            path = link.file_path
            logger.info('Processing %s', display_path(path))
        else:
            logger.info('Collecting %s', req.req or req)

        with indent_log():
            # @@ if filesystem packages are not marked
            # editable in a req, a non deterministic error
            # occurs when the script attempts to unpack the
            # build directory
            # Since source_dir is only set for editable requirements.
            assert req.source_dir is None
            req.ensure_has_source_dir(self.build_dir)
            # If a checkout exists, it's unwise to keep going.  version
            # inconsistencies are logged later, but do not fail the
            # installation.
            # FIXME: this won't upgrade when there's an existing
            # package unpacked in `req.source_dir`
            if os.path.exists(os.path.join(req.source_dir, 'setup.py')):
                raise PreviousBuildDirError(
                    "pip can't proceed with requirements '{}' due to a"
                    " pre-existing build directory ({}). This is "
                    "likely due to a previous installation that failed"
                    ". pip is being responsible and not assuming it "
                    "can delete this. Please delete it and try again."
                    .format(req, req.source_dir)
                )

            # Now that we have the real link, we can tell what kind of
            # requirements we have and raise some more informative errors
            # than otherwise. (For example, we can raise VcsHashUnsupported
            # for a VCS URL rather than HashMissing.)
            if self.require_hashes:
                # We could check these first 2 conditions inside
                # unpack_url and save repetition of conditions, but then
                # we would report less-useful error messages for
                # unhashable requirements, complaining that there's no
                # hash provided.
                if link.is_vcs:
                    raise VcsHashUnsupported()
                elif link.is_existing_dir():
                    raise DirectoryUrlHashUnsupported()
                if not req.original_link and not req.is_pinned:
                    # Unpinned packages are asking for trouble when a new
                    # version is uploaded. This isn't a security check, but
                    # it saves users a surprising hash mismatch in the
                    # future.
                    #
                    # file:/// URLs aren't pinnable, so don't complain
                    # about them not being pinned.
                    raise HashUnpinned()

            hashes = req.hashes(trust_internet=not self.require_hashes)
            if self.require_hashes and not hashes:
                # Known-good hashes are missing for this requirement, so
                # shim it with a facade object that will provoke hash
                # computation and then raise a HashMissing exception
                # showing the user what the hash should be.
                hashes = MissingHashes()

            download_dir = self.download_dir
            if link.is_wheel and self.wheel_download_dir:
                # when doing 'pip wheel` we download wheels to a
                # dedicated dir.
                download_dir = self.wheel_download_dir

            try:
                local_path = unpack_url(
                    link, req.source_dir, self.downloader, download_dir,
                    hashes=hashes,
                )
            except requests.HTTPError as exc:
                logger.critical(
                    'Could not install requirement %s because of error %s',
                    req,
                    exc,
                )
                raise InstallationError(
                    'Could not install requirement {} because of HTTP '
                    'error {} for URL {}'.format(req, exc, link)
                )

            # For use in later processing, preserve the file path on the
            # requirement.
            if local_path:
                req.local_file_path = local_path

            if link.is_wheel:
                if download_dir:
                    # When downloading, we only unpack wheels to get
                    # metadata.
                    autodelete_unpacked = True
                else:
                    # When installing a wheel, we use the unpacked
                    # wheel.
                    autodelete_unpacked = False
            else:
                # We always delete unpacked sdists after pip runs.
                autodelete_unpacked = True
            if autodelete_unpacked:
                write_delete_marker_file(req.source_dir)

            abstract_dist = _get_prepared_distribution(
                req, self.req_tracker, self.finder, self.build_isolation,
            )

            if download_dir:
                if link.is_existing_dir():
                    logger.info('Link is a directory, ignoring download_dir')
                elif local_path and not os.path.exists(
                    os.path.join(download_dir, link.filename)
                ):
                    _copy_file(local_path, download_dir, link)

            if self._download_should_save:
                # Make a .zip of the source_dir we already created.
                if link.is_vcs:
                    req.archive(self.download_dir)
        return abstract_dist

    def prepare_editable_requirement(
        self,
        req,  # type: InstallRequirement
    ):
        # type: (...) -> AbstractDistribution
        """Prepare an editable requirement
        """
        assert req.editable, "cannot prepare a non-editable req as editable"

        logger.info('Obtaining %s', req)

        with indent_log():
            if self.require_hashes:
                raise InstallationError(
                    'The editable requirement {} cannot be installed when '
                    'requiring hashes, because there is no single file to '
                    'hash.'.format(req)
                )
            req.ensure_has_source_dir(self.src_dir)
            req.update_editable(not self._download_should_save)

            abstract_dist = _get_prepared_distribution(
                req, self.req_tracker, self.finder, self.build_isolation,
            )

            if self._download_should_save:
                req.archive(self.download_dir)
            req.check_if_exists(self.use_user_site)

        return abstract_dist

    def prepare_installed_requirement(
        self,
        req,  # type: InstallRequirement
        skip_reason  # type: str
    ):
        # type: (...) -> AbstractDistribution
        """Prepare an already-installed requirement
        """
        assert req.satisfied_by, "req should have been satisfied but isn't"
        assert skip_reason is not None, (
            "did not get skip reason skipped but req.satisfied_by "
            "is set to {}".format(req.satisfied_by)
        )
        logger.info(
            'Requirement %s: %s (%s)',
            skip_reason, req, req.satisfied_by.version
        )
        with indent_log():
            if self.require_hashes:
                logger.debug(
                    'Since it is already installed, we are trusting this '
                    'package without checking its hash. To ensure a '
                    'completely repeatable environment, install into an '
                    'empty virtualenv.'
                )
            abstract_dist = InstalledDistribution(req)

        return abstract_dist



from .. import Provider as CompanyProvider


class Provider(CompanyProvider):

    formats = (
        '{{first_name}} և {{first_name}} {{company_suffix}}',
        '{{last_name}} {{company_suffix}}',
        '{{last_name}} և {{last_name}} {{company_suffix}}'
        '{{last_name}}, {{last_name}} և {{last_name}} {{company_suffix}}',
    )

    company_suffixes = ('ՍՊԸ', 'ՀՁ', 'ՓԲԸ', 'ԲԲԸ', 'Գրուպ', 'Հոլդինգ')

    catch_phrase_words = (
        ('առաջավոր',
         'բարելավված',
         'ավտոմատացված',
         'հավասարակշռված',
         'կենտրոնացված',
         'համատեղելի',
         'կարգավորելի',
         'անհատականացված',
         'ապակենտրոնացված',
         'թվայնացված',
         'տարածված',
         'փոքրացված',
         'ընդլայնված',
         'էրգոնիկ',
         'բացառիկ',
         'երկարացված',
         'լիովին կոնֆիգուրացվող',
         'ֆունկցիոնալ հիմունքներով',
         'հիմնական',
         'հորիզոնական',
         'իրականացված',
         'նորարական',
         'ինտեգրված',
         'ինտուիտիվ',
         'պարտադիր',
         'բազմուղի',
         'բազմաշերտ',
         'ցանցային',
         'բաց կոդով',
         'օպերատիվ',
         'օպտիմալացված',
         'օրգանական',
         'կազմակերպված',
         'կայուն',
         'կենսունակ',
         'ավարտված',
         'բևեռացված',
         'կանխարգելող',
         'ակտիվ',
         'ծրագրավորելի',
         'առաջադիմական',
         'որակով',
         'ռեակտիվ',
         'իրականացված',
         'նվազեցված',
         'դիմացկուն',
         'անխափան',
         'ապահով',
         'համատեղելի',
         'հեշտացված',
         'փոխարկելի',
         'սինխրոնիզացված',
         'ունիվերսալ',
         'ճկուն',
         'վիրտուալ'),
        ('3-րդ սերնդի',
         '4-րդ սերնդի',
         '5-րդ սերնդի',
         '6-րդ սերնդի',
         'ասիմետրիկ',
         'ասինխրոն',
         'թողունակությունը վերահսկվող',
         'երկկողմանի',
         'հստակ մտածող',
         'համաձայնեցված',
         'բաղադրյալ',
         'դիդակտիկ',
         'ուղղորդիչ',
         'դիսկրետ',
         'բացահայտ',
         'գլոբալ',
         'բարձր մակարդակի',
         'ամբողջական',
         'միատարր',
         'հիբրիդ',
         'ազդեցիկ',
         'ինտերակտիվ',
         'միջանկյալ',
         'առաջատար',
         'տեղային',
         'լոգիստիկ',
         'սիստեմատիկ',
         'մոդուլային',
         'չեզոք',
         'հաջորդ սերնդի',
         'օբյեկտի վրա հիմնված',
         'օպտիմալ',
         'արմատական',
         'փոխադարձ',
         'ռեգիոնալ',
         'երկրորդական',
         'կայուն',
         'ստատիկ',
         'համակարգված',
         'համակարգային',
         'շոշափելի',
         'երրորդական',
         'անցումային',
         'միատեսակ',
         'լավ մոդուլացված',
         'առանց թերությունների'),
        ('կարողություն',
         'մուտք',
         'ադապտեր',
         'ալգորիթմ',
         'միություն',
         'վերլուծիչ',
         'ծրագրային ապահովում',
         'մոտեցում',
         'արխիվ',
         'արհեստական բանականություն',
         'վերաբերմունք',
         'ընդունակություն',
         'կարողություն',
         'մարտահրավեր',
         'գործակցություն',
         'բարդություն',
         'գաղափար',
         'համախմբվածություն',
         'տվյալների բազա',
         'տվյալների պահեստ',
         'սահմանում',
         'իմացություն',
         'կոդավորում',
         'գաղտնագրում',
         'կանխատեսում',
         'հենքային ծրագիր',
         'ֆունկցիա',
         'գործառույթ',
         'գրաֆիկական ինտերֆեյս',
         'սարքային ապահովում',
         'հիերարխիա',
         'հանգույց',
         'ենթակառուցվածք',
         'նախաձեռնություն',
         'ծրագրի ներդրում',
         'հրահանգների հավաքածու',
         'ինտերֆեյս',
         'ինտրանետ',
         'գիտելիքների բազա',
         'տեղական ցանց',
         'մատրիցա',
         'մեթոդաբանություն',
         'միջանկյալ շերտ',
         'միգրացիա',
         'մոդել',
         'կարգավորիչ',
         'մոնիտորինգ',
         'բաց համակարգ',
         'պարադիգմ',
         'պորտալ',
         'գնային կառուցվածք',
         'արդյունավետություն',
         'նախագիծ',
         'ապահովված գիծ',
         'ծրագրային ապահովում',
         'լուծում',
         'ստանդարտացում',
         'ստրատեգիա',
         'կառուցվածք',
         'օպերատիվ խումբ',
         'արտադրողականություն',
         'ժամանակացույց',
         'գործիք',
         'օգտագործում',
         'կայք',
         'աշխատուժ'))

    bsWords = (
        ('իրականացնել',
         'օգտագործել',
         'ինտեգրել',
         'ռացիոնալացնել',
         'օպտիմալացնել',
         'փոխակերպել',
         'ընդգրկել',
         'ակտիվացնել',
         'կազմակերպել',
         'նախագծել',
         'խթանել',
         'ձևափոխել',
         'արտոնել',
         'դրամայնացնել',
         'հեշտացնել',
         'վերցնել',
         'աճեցնել',
         'սինթեզել',
         'առաքել',
         'զբաղվել',
         'առավելագույնի հասցնել',
         'արագացնել',
         'միջնորդել',
         'պատկերացնել',
         'վերափոխել',
         'ընդլայնել',
         'նախաձեռնել',
         'հեղափոխականացնել',
         'առաջացնել',
         'օգտագործել',
         'զարգացնել',
         'արտադրանքի վերածել'),
        ('ուղղահայաց',
         'ակտիվ',
         'դիմացկուն',
         'հեղափոխական',
         'առաջատար',
         'նորարարական',
         'ինտուիտիվ',
         'ռազմավարական',
         'էլեկտրոնային',
         'գլոբալ',
         'վիրտուալ',
         'դինամիկ',
         'գրավիչ',
         'ինտերակտիվ',
         'արդյունավետ',
         'ընդարձակելի',
         'պատրաստի',
         'ինտեգրված',
         'ազդեցիկ',
         'անլար',
         'թափանցիկ',
         'հաջորդ սերնդի',
         'ժամանակակից',
         'հարմարեցված',
         'համատարած',
         'ազդեցիկ',
         'ամբողջական',
         'հարուստ',
         'անվճար'),
        ('պարադիգմներ',
         'շուկաներ',
         'ենթակառուցվածքներ',
         'պլատֆորմներ',
         'նախաձեռնություններ',
         'ուղիներ',
         'համայնքներ',
         'լուծումներ',
         'պորտալներ',
         'տեխնոլոգիաներ',
         'հարաբերություններ',
         'կառուցվածքներ',
         'ինտերֆեյսներ',
         'շուկաներ',
         'համակարգեր',
         'մոդելներ',
         'օգտագործողներ',
         'սխեմաներ',
         'ցանցեր',
         'ծրագրեր',
         'չափանիշներ',
         'բիզնես',
         'գործառույթներ',
         'փորձառություններ',
         'մեթոդաբանություններ'))
